"""
Model for a Zwave Node's device config.

https://zwave-js.github.io/node-zwave-js/#/api/node?id=deviceconfig
"""

from __future__ import annotations

from typing import Any, Literal, TypedDict


class DeviceDeviceDataType(TypedDict, total=False):
    """Represent a device device data dict type."""

    productType: str | int
    productId: str | int


class DeviceDevice:
    """Model for a Zwave Node's device config's device."""

    def __init__(self, data: DeviceDeviceDataType) -> None:
        """Initialize."""
        self.data = data

    @property
    def product_type(self) -> str | int | None:
        """Return product type."""
        return self.data.get("productType")

    @property
    def product_id(self) -> str | int | None:
        """Return product id."""
        return self.data.get("productId")


class DeviceFirmwareVersionRangeDataType(TypedDict, total=False):
    """Represent a device firmware version range data dict type."""

    min: str
    max: str


class DeviceFirmwareVersionRange:
    """Model for a Zwave Node's device config's firmware version range."""

    def __init__(self, data: DeviceFirmwareVersionRangeDataType) -> None:
        """Initialize."""
        self.data = data

    @property
    def min(self) -> str | None:
        """Return min version."""
        return self.data.get("min")

    @property
    def max(self) -> str | None:
        """Return max version."""
        return self.data.get("max")


class CommentDataType(TypedDict):
    """Represent a device config's comment data dict type."""

    # See PR for suggested meanings of each level:
    # https://github.com/zwave-js/node-zwave-js/pull/3947
    level: Literal["info", "warning", "error"]
    text: str


class DeviceMetadataDataType(TypedDict, total=False):
    """Represent a device metadata data dict type."""

    wakeup: str
    inclusion: str
    exclusion: str
    reset: str
    manual: str
    comments: CommentDataType | list[CommentDataType]


class DeviceMetadata:
    """Model for a Zwave Node's device config's metadata."""

    def __init__(self, data: DeviceMetadataDataType) -> None:
        """Initialize."""
        self.data = data

    @property
    def wakeup(self) -> str | None:
        """Return wakeup instructions."""
        return self.data.get("wakeup")

    @property
    def inclusion(self) -> str | None:
        """Return inclusion instructions."""
        return self.data.get("inclusion")

    @property
    def exclusion(self) -> str | None:
        """Return exclusion instructions."""
        return self.data.get("exclusion")

    @property
    def reset(self) -> str | None:
        """Return reset instructions."""
        return self.data.get("reset")

    @property
    def manual(self) -> str | None:
        """Return manual instructions."""
        return self.data.get("manual")

    @property
    def comments(self) -> list[CommentDataType]:
        """Return list of comments about device."""
        comments = self.data.get("comments", [])
        if isinstance(comments, dict):
            return [comments]
        return comments


class DeviceConfigDataType(TypedDict, total=False):
    """Represent a device config data dict type."""

    filename: str
    manufacturer: str
    manufacturerId: int
    label: str
    description: str
    devices: list[DeviceDeviceDataType]
    firmwareVersion: DeviceFirmwareVersionRangeDataType
    associations: dict[str, dict]
    paramInformation: dict[str, dict]
    supportsZWavePlus: bool
    proprietary: dict
    compat: dict[str, Any]
    metadata: DeviceMetadataDataType
    isEmbedded: bool


class DeviceConfig:
    """Model for a Zwave Node's device config."""

    def __init__(self, data: DeviceConfigDataType) -> None:
        """Initialize."""
        self.data = data
        self._devices = [
            DeviceDevice(device) for device in self.data.get("devices", [])
        ]
        self._firmware_version = DeviceFirmwareVersionRange(
            self.data.get("firmwareVersion", {})
        )
        self._metadata = DeviceMetadata(self.data.get("metadata", {}))

    @property
    def filename(self) -> str | None:
        """Return config filename."""
        return self.data.get("filename")

    @property
    def manufacturer(self) -> str | None:
        """Return name of the manufacturer."""
        return self.data.get("manufacturer")

    @property
    def manufacturer_id(self) -> int | None:
        """Return manufacturer id (as defined in specs)."""
        return self.data.get("manufacturerId")

    @property
    def label(self) -> str | None:
        """Return short label for the device."""
        return self.data.get("label")

    @property
    def description(self) -> str | None:
        """Return longer description of the device, usually the full name."""
        return self.data.get("description")

    @property
    def devices(self) -> list[DeviceDevice]:
        """Return list of product type and product ID combinations."""
        return self._devices

    @property
    def firmware_version(self) -> DeviceFirmwareVersionRange:
        """Return firmware version range this config is valid for."""
        return self._firmware_version

    @property
    def associations(self) -> dict[str, dict]:
        """Return dict of association groups the device supports."""
        return self.data.get("associations", {})

    @property
    def param_information(self) -> dict[str, dict]:
        """Return dictionary of configuration parameters the device supports."""
        return self.data.get("paramInformation", {})

    @property
    def supports_zwave_plus(self) -> bool | None:
        """Return if the device complies with the Z-Wave+ standard."""
        return self.data.get("supportsZWavePlus")

    @property
    def proprietary(self) -> dict:
        """Return dictionary of settings for the proprietary CC."""
        return self.data.get("proprietary", {})

    @property
    def compat(self) -> dict[str, dict]:
        """Return compatibility flags."""
        return self.data.get("compat", {})

    @property
    def metadata(self) -> DeviceMetadata:
        """Return metadata."""
        return self._metadata

    @property
    def is_embedded(self) -> bool | None:
        """Return whether device config is embedded in zwave-js-server."""
        return self.data.get("isEmbedded")

    def to_dict(self) -> DeviceConfigDataType:
        """Return dict representation of device config."""
        return self.data.copy()



"""Test the driver model."""

import json
from typing import Any

import pytest

from zwave_js_server.const import LogLevel
from zwave_js_server.event import Event
from zwave_js_server.model import (
    log_config as log_config_pkg,
    log_message as log_message_pkg,
)
from zwave_js_server.model.driver import Driver
from zwave_js_server.model.driver.firmware import DriverFirmwareUpdateStatus
from zwave_js_server.model.firmware import FirmwareUpdateData, FirmwareUpdateInfo

from .. import load_fixture
from ..common import MockCommandProtocol
from .common import FIRMWARE_UPDATE_INFO


def test_from_state(client, log_config):
    """Test from_state method."""
    ws_msgs = load_fixture("basic_dump.txt").strip().split("\n")

    driver = Driver(client, json.loads(ws_msgs[0])["result"]["state"], log_config)
    assert driver == Driver(
        client, json.loads(ws_msgs[0])["result"]["state"], log_config
    )
    assert driver != driver.controller.home_id
    assert hash(driver) == hash(driver.controller.home_id)

    for msg in ws_msgs[1:]:
        msg = json.loads(msg)
        assert msg["type"] == "event"
        event = Event(type=msg["event"]["event"], data=msg["event"])
        driver.receive_event(event)

    assert len(driver.controller.nodes) == 8


async def test_update_log_config(driver, uuid4, mock_command):
    """Test update log config."""
    # Update log level
    ack_commands = mock_command(
        {"command": "driver.update_log_config", "config": {"level": "error"}},
        {"success": True},
    )

    assert (
        await driver.async_update_log_config(
            log_config_pkg.LogConfig(level=LogLevel.ERROR)
        )
        is None
    )

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.update_log_config",
        "config": {"level": "error"},
        "messageId": uuid4,
    }

    # Update all parameters
    ack_commands = mock_command(
        {
            "command": "driver.update_log_config",
            "config": {
                "enabled": True,
                "level": "error",
                "logToFile": True,
                "filename": "/test.txt",
                "forceConsole": True,
            },
        },
        {"success": True},
    )
    assert (
        await driver.async_update_log_config(
            log_config_pkg.LogConfig(
                enabled=True,
                level=LogLevel.ERROR,
                log_to_file=True,
                filename="/test.txt",
                force_console=True,
            )
        )
        is None
    )

    assert len(ack_commands) == 2
    assert ack_commands[1] == {
        "command": "driver.update_log_config",
        "config": {
            "enabled": True,
            "level": "error",
            "logToFile": True,
            "filename": "/test.txt",
            "forceConsole": True,
        },
        "messageId": uuid4,
    }


async def test_get_log_config(driver, uuid4, mock_command):
    """Test set value."""
    ack_commands = mock_command(
        {"command": "driver.get_log_config"},
        {
            "success": True,
            "config": {
                "enabled": True,
                "level": "error",
                "logToFile": False,
                "filename": "/test.txt",
                "forceConsole": False,
            },
        },
    )
    log_config = await driver.async_get_log_config()

    assert len(ack_commands) == 1
    assert ack_commands[0] == {"command": "driver.get_log_config", "messageId": uuid4}

    assert log_config.enabled
    assert log_config.level == LogLevel.ERROR
    assert log_config.log_to_file is False
    assert log_config.filename == "/test.txt"
    assert log_config.force_console is False


async def test_listening_logs(client, driver, uuid4, mock_command):
    """Test listening to logs helpers."""
    # Test that start listening to logs command is sent
    ack_commands = mock_command(
        {"command": "start_listening_logs"},
        {"success": True},
    )
    await client.async_start_listening_logs()

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "start_listening_logs",
        "messageId": uuid4,
    }

    # Test that stop listening to logs command is sent
    ack_commands = mock_command(
        {"command": "stop_listening_logs"},
        {"success": True},
    )
    await client.async_stop_listening_logs()

    assert len(ack_commands) == 2
    assert ack_commands[1] == {
        "command": "stop_listening_logs",
        "messageId": uuid4,
    }

    # Test receiving a log message event
    event = Event(
        type="logging",
        data={
            "source": "driver",
            "event": "logging",
            "formattedMessage": [
                "2021-04-18T18:03:34.051Z CNTRLR   [Node 005] [~] \n",
                "test",
            ],
            "level": "debug",
            "primaryTags": "[Node 005] [~] [Notification]",
            "secondaryTags": "[Endpoint 0]",
            "message": "Home Security[Motion sensor status]\n: 8 => 0",
            "direction": "  ",
            "label": "CNTRLR",
            "timestamp": "2021-04-18T18:03:34.051Z",
            "multiline": True,
            "secondaryTagPadding": -1,
            "context": {
                "source": "controller",
                "type": "node",
                "nodeId": 5,
                "header": "Notification",
                "direction": "none",
                "change": "notification",
                "endpoint": 0,
            },
        },
    )
    driver.receive_event(event)
    assert "log_message" in event.data
    assert isinstance(event.data["log_message"], log_message_pkg.LogMessage)
    log_message = event.data["log_message"]
    assert log_message.message == ["Home Security[Motion sensor status]", ": 8 => 0"]
    assert log_message.formatted_message == [
        "2021-04-18T18:03:34.051Z CNTRLR   [Node 005] [~] ",
        "test",
    ]
    assert log_message.direction == "  "
    assert log_message.primary_tags == "[Node 005] [~] [Notification]"
    assert log_message.secondary_tags == "[Endpoint 0]"
    assert log_message.level == "debug"
    assert log_message.label == "CNTRLR"
    assert log_message.multiline
    assert log_message.secondary_tag_padding == -1
    assert log_message.timestamp == "2021-04-18T18:03:34.051Z"

    context = log_message.context
    assert context.change == "notification"
    assert context.direction == "none"
    assert context.endpoint == 0
    assert context.header == "Notification"
    assert context.node_id == 5
    assert context.source == "controller"
    assert context.type == "node"
    assert context.internal is None
    assert context.property_ is None
    assert context.property_key is None
    assert context.command_class is None


async def test_statistics(driver, uuid4, mock_command):
    """Test statistics commands."""
    # Test that enable_statistics command is sent
    ack_commands = mock_command(
        {
            "command": "driver.enable_statistics",
            "applicationName": "test_name",
            "applicationVersion": "test_version",
        },
        {"success": True},
    )
    await driver.async_enable_statistics("test_name", "test_version")

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.enable_statistics",
        "applicationName": "test_name",
        "applicationVersion": "test_version",
        "messageId": uuid4,
    }

    # Test that disable_statistics command is sent
    ack_commands = mock_command(
        {"command": "driver.disable_statistics"},
        {"success": True},
    )
    await driver.async_disable_statistics()

    assert len(ack_commands) == 2
    assert ack_commands[1] == {
        "command": "driver.disable_statistics",
        "messageId": uuid4,
    }

    # Test that is statistics_enabled command is sent
    ack_commands = mock_command(
        {"command": "driver.is_statistics_enabled"},
        {"success": True, "statisticsEnabled": True},
    )
    assert await driver.async_is_statistics_enabled()

    assert len(ack_commands) == 3
    assert ack_commands[2] == {
        "command": "driver.is_statistics_enabled",
        "messageId": uuid4,
    }


async def test_log_config_updated(driver):
    """Test the log_config_updated event."""
    # Modify current log config in an update and assert that it changed
    assert driver.log_config.level != LogLevel.SILLY
    log_config = driver.log_config.to_dict()
    log_config["level"] = "silly"
    event = Event(
        "log config updated",
        data={"source": "driver", "event": "log config updated", "config": log_config},
    )
    driver.receive_event(event)
    assert driver.log_config.level == LogLevel.SILLY


async def test_check_for_config_updates(driver, uuid4, mock_command):
    """Test driver.check_for_config_updates command."""
    ack_commands = mock_command(
        {"command": "driver.check_for_config_updates"},
        {"updateAvailable": False, "installedVersion": "1.0.0"},
    )

    check_config_updates = await driver.async_check_for_config_updates()
    assert check_config_updates.installed_version == "1.0.0"
    assert check_config_updates.update_available is False
    assert check_config_updates.new_version is None

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.check_for_config_updates",
        "messageId": uuid4,
    }


async def test_install_config_update(driver, uuid4, mock_command):
    """Test driver.install_config_update command."""
    ack_commands = mock_command(
        {"command": "driver.install_config_update"},
        {"success": False},
    )

    assert not await driver.async_install_config_update()

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.install_config_update",
        "messageId": uuid4,
    }


async def test_set_preferred_scales(driver, uuid4, mock_command):
    """Test driver.set_preferred_scales command."""
    ack_commands = mock_command({"command": "driver.set_preferred_scales"}, {})

    assert not await driver.async_set_preferred_scales({1: 1})

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.set_preferred_scales",
        "scales": {1: 1},
        "messageId": uuid4,
    }


async def test_hard_reset(driver, uuid4, mock_command):
    """Test driver hard reset command."""
    ack_commands = mock_command({"command": "driver.hard_reset"}, {})

    assert not await driver.async_hard_reset()

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.hard_reset",
        "messageId": uuid4,
    }


async def test_try_soft_reset(driver, uuid4, mock_command):
    """Test driver try soft reset command."""
    ack_commands = mock_command({"command": "driver.try_soft_reset"}, {})

    assert not await driver.async_try_soft_reset()

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.try_soft_reset",
        "messageId": uuid4,
    }


async def test_soft_reset(driver, uuid4, mock_command):
    """Test driver soft reset command."""
    ack_commands = mock_command({"command": "driver.soft_reset"}, {})

    assert not await driver.async_soft_reset()

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.soft_reset",
        "messageId": uuid4,
    }


async def test_shutdown(driver, uuid4, mock_command):
    """Test driver shutdown command."""
    ack_commands = mock_command({"command": "driver.shutdown"}, {"success": True})

    assert await driver.async_shutdown()

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.shutdown",
        "messageId": uuid4,
    }


async def test_unknown_event(driver):
    """Test that an unknown event type causes an exception."""
    with pytest.raises(KeyError):
        assert driver.receive_event(Event("unknown_event", {"source": "driver"}))


async def test_all_nodes_ready_event(driver):
    """Test that the all nodes ready event is succesfully validated by pydantic."""
    event = Event("all nodes ready", {"source": "driver", "event": "all nodes ready"})
    driver.receive_event(event)


async def test_driver_ready_event(driver):
    """Test that the driver ready event is succesfully validated by pydantic."""
    event_type = "driver ready"
    event_data = {"source": "driver", "event": event_type}
    event = Event(event_type, event_data)

    def callback(data: dict) -> None:
        assert data == event_data

    driver.on(event_type, callback)
    driver.receive_event(event)


def test_config_manager(driver):
    """Test the driver has the config manager property."""
    assert driver.config_manager is not None
    assert driver.config_manager._client is driver.client


async def test_firmware_events(driver):
    """Test firmware events."""
    assert driver.firmware_update_progress is None
    event = Event(
        type="firmware update progress",
        data={
            "source": "driver",
            "event": "firmware update progress",
            "progress": {
                "sentFragments": 1,
                "totalFragments": 10,
                "progress": 10.0,
            },
        },
    )

    driver.receive_event(event)
    progress = event.data["firmware_update_progress"]
    assert progress.sent_fragments == 1
    assert progress.total_fragments == 10
    assert progress.progress == 10.0
    assert driver.firmware_update_progress
    assert driver.firmware_update_progress.sent_fragments == 1
    assert driver.firmware_update_progress.total_fragments == 10
    assert driver.firmware_update_progress.progress == 10.0

    event = Event(
        type="firmware update finished",
        data={
            "source": "driver",
            "event": "firmware update finished",
            "result": {"status": 255, "success": True},
        },
    )

    driver.receive_event(event)
    result = event.data["firmware_update_finished"]
    assert result.status == DriverFirmwareUpdateStatus.OK
    assert result.success
    assert driver.firmware_update_progress is None


@pytest.mark.parametrize("progress", [True, False])
async def test_is_otw_firmware_update_in_progress(
    driver: Driver,
    uuid4: str,
    mock_command: MockCommandProtocol,
    progress: bool,
) -> None:
    """Test is_otw_firmware_update_in_progress command."""
    ack_commands = mock_command(
        {"command": "driver.is_otw_firmware_update_in_progress"},
        {"progress": progress},
    )
    assert await driver.async_is_otw_firmware_update_in_progress() is progress

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.is_otw_firmware_update_in_progress",
        "messageId": uuid4,
    }


@pytest.mark.parametrize(
    ("firmware_update_param", "expected_params"),
    [
        (
            {
                "update_data": FirmwareUpdateData(
                    filename="test-file", file=b"test", file_format=None
                )
            },
            FirmwareUpdateData(
                filename="test-file", file=b"test", file_format=None
            ).to_dict(),
        ),
        (
            {"update_info": FirmwareUpdateInfo.from_dict(FIRMWARE_UPDATE_INFO)},
            {"updateInfo": FIRMWARE_UPDATE_INFO},
        ),
    ],
)
@pytest.mark.parametrize(("status", "success"), [(1, True), (0, False)])
async def test_firmware_update_otw(
    driver: Driver,
    uuid4: str,
    mock_command: MockCommandProtocol,
    status: int,
    success: bool,
    firmware_update_param: dict[str, Any],
    expected_params: dict[str, Any],
) -> None:
    """Test firmware_update_otw command."""
    ack_commands = mock_command(
        {"command": "driver.firmware_update_otw", **expected_params},
        {"result": {"status": status, "success": success}},
    )
    result = await driver.async_firmware_update_otw(**firmware_update_param)

    assert result.status == status
    assert result.success is success

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "driver.firmware_update_otw",
        "messageId": uuid4,
        **expected_params,
    }


@pytest.mark.parametrize(
    "firmware_update_param",
    [
        {},  # No parameters
        {
            "update_data": FirmwareUpdateData(
                filename="test-file", file=b"test", file_format=None
            ),
            "update_info": FirmwareUpdateInfo.from_dict(FIRMWARE_UPDATE_INFO),
        },  # Invalid parameters
    ],
)
async def test_firmware_update_otw_invalid_params(
    driver: Driver,
    firmware_update_param: dict[str, FirmwareUpdateData | FirmwareUpdateInfo],
) -> None:
    """Test firmware_update_otw command invalid parameters."""
    with pytest.raises(ValueError):
        # Should raise ValueError if no parameters are provided or both are provided
        await driver.async_firmware_update_otw(**firmware_update_param)



"""Constants for the Multilevel Sensor CC."""

# ----------------------------------------------------------------------------------- #
# **BEGINNING OF AUTOGENERATED CONTENT** (TO ADD ADDITIONAL MANUAL CONTENT, LOOK FOR  #
# THE "END OF AUTOGENERATED CONTENT" COMMENT BLOCK AND ADD YOUR CODE BELOW IT)        #
# ----------------------------------------------------------------------------------- #

from __future__ import annotations

from enum import IntEnum

CC_SPECIFIC_SCALE = "scale"
CC_SPECIFIC_SENSOR_TYPE = "sensorType"

ACCELERATION_X_AXIS_PROPERTY = "Acceleration X-axis"
ACCELERATION_Y_AXIS_PROPERTY = "Acceleration Y-axis"
ACCELERATION_Z_AXIS_PROPERTY = "Acceleration Z-axis"
AIR_FLOW_PROPERTY = "Air flow"
AIR_TEMPERATURE_PROPERTY = "Air temperature"
AMMONIA_PROPERTY = "Ammonia"
ANGLE_POSITION_PROPERTY = "Angle position"
APPLIED_FORCE_ON_THE_SENSOR_PROPERTY = "Applied force on the sensor"
ATMOSPHERIC_PRESSURE_PROPERTY = "Atmospheric pressure"
BAROMETRIC_PRESSURE_PROPERTY = "Barometric pressure"
BASIS_METABOLIC_RATE_PROPERTY = "Basis metabolic rate (BMR)"
BLOOD_PRESSURE_PROPERTY = "Blood pressure"
BODY_MASS_INDEX_PROPERTY = "Body Mass Index (BMI)"
BOILER_WATER_TEMPERATURE_PROPERTY = "Boiler water temperature"
BONE_MASS_PROPERTY = "Bone mass"
CARBON_DIOXIDE_LEVEL_PROPERTY = "Carbon dioxide (CO₂) level"
CARBON_MONOXIDE_LEVEL_PROPERTY = "Carbon monoxide (CO) level"
CONDENSER_COIL_TEMPERATURE_PROPERTY = "Condenser Coil temperature"
CURRENT_PROPERTY = "Current"
DEFROST_TEMPERATURE_PROPERTY = "Defrost temperature"
DEW_POINT_PROPERTY = "Dew point"
DIRECTION_PROPERTY = "Direction"
DISCHARGE_LINE_TEMPERATURE_PROPERTY = "Discharge Line temperature"
DISCHARGE_PRESSURE_PROPERTY = "Discharge Pressure"
DISTANCE_PROPERTY = "Distance"
DOMESTIC_HOT_WATER_TEMPERATURE_PROPERTY = "Domestic Hot Water (DHW) temperature"
ELECTRICAL_CONDUCTIVITY_PROPERTY = "Electrical conductivity"
ELECTRICAL_RESISTIVITY_PROPERTY = "Electrical resistivity"
EVAPORATOR_COIL_TEMPERATURE_PROPERTY = "Evaporator Coil temperature"
EXHAUST_TEMPERATURE_PROPERTY = "Exhaust temperature"
FAT_MASS_PROPERTY = "Fat mass"
FORMALDEHYDE_LEVEL_PROPERTY = "Formaldehyde (CH₂O) level"
FREQUENCY_PROPERTY = "Frequency"
GENERAL_PURPOSE_PROPERTY = "General purpose"
HEART_RATE_PROPERTY = "Heart rate"
HEART_RATE_LF_HF_RATIO_PROPERTY = "Heart Rate LF/HF ratio"
HUMIDITY_PROPERTY = "Humidity"
ILLUMINANCE_PROPERTY = "Illuminance"
LEAD_PROPERTY = "Lead"
LIQUID_LINE_TEMPERATURE_PROPERTY = "Liquid Line temperature"
LOUDNESS_PROPERTY = "Loudness"
METHANE_DENSITY_PROPERTY = "Methane (CH₄) density"
MOISTURE_PROPERTY = "Moisture"
MOTION_DIRECTION_PROPERTY = "Motion Direction"
MUSCLE_MASS_PROPERTY = "Muscle mass"
NITROGEN_DIOXIDE_PROPERTY = "Nitrogen dioxide"
OUTSIDE_TEMPERATURE_PROPERTY = "Outside temperature"
OZONE_PROPERTY = "Ozone"
PARTICULATE_MATTER_1_PROPERTY = "Particulate Matter 1"
PARTICULATE_MATTER_10_PROPERTY = "Particulate Matter 10"
PARTICULATE_MATTER_2_5_PROPERTY = "Particulate Matter 2.5"
PERSON_COUNTER_ENTERING_PROPERTY = "Person counter (entering)"
PERSON_COUNTER_EXITING_PROPERTY = "Person counter (exiting)"
POWER_PROPERTY = "Power"
RADON_CONCENTRATION_PROPERTY = "Radon concentration"
RAIN_RATE_PROPERTY = "Rain rate"
RELATIVE_MODULATION_LEVEL_PROPERTY = "Relative Modulation level"
RESPIRATORY_RATE_PROPERTY = "Respiratory rate"
RETURN_AIR_TEMPERATURE_PROPERTY = "Return Air temperature"
RF_SIGNAL_STRENGTH_PROPERTY = "RF signal strength"
ROTATION_PROPERTY = "Rotation"
SEISMIC_INTENSITY_PROPERTY = "Seismic Intensity"
SEISMIC_MAGNITUDE_PROPERTY = "Seismic magnitude"
SMOKE_DENSITY_PROPERTY = "Smoke density"
SOIL_HUMIDITY_PROPERTY = "Soil humidity"
SOIL_REACTIVITY_PROPERTY = "Soil reactivity"
SOIL_SALINITY_PROPERTY = "Soil salinity"
SOIL_TEMPERATURE_PROPERTY = "Soil temperature"
SOLAR_RADIATION_PROPERTY = "Solar radiation"
SUCTION_PRESSURE_PROPERTY = "Suction Pressure"
SULFUR_DIOXIDE_PROPERTY = "Sulfur dioxide"
SUPPLY_AIR_TEMPERATURE_PROPERTY = "Supply Air temperature"
TANK_CAPACITY_PROPERTY = "Tank capacity"
TARGET_TEMPERATURE_PROPERTY = "Target temperature"
TIDE_LEVEL_PROPERTY = "Tide level"
TIME_PROPERTY = "Time"
TOTAL_BODY_WATER_PROPERTY = "Total body water (TBW)"
ULTRAVIOLET_PROPERTY = "Ultraviolet"
VELOCITY_PROPERTY = "Velocity"
VOLATILE_ORGANIC_COMPOUND_LEVEL_PROPERTY = "Volatile Organic Compound level"
VOLTAGE_PROPERTY = "Voltage"
WATER_ACIDITY_PROPERTY = "Water acidity"
WATER_CHLORINE_LEVEL_PROPERTY = "Water Chlorine level"
WATER_FLOW_PROPERTY = "Water flow"
WATER_OXIDATION_REDUCTION_POTENTIAL_PROPERTY = "Water Oxidation reduction potential"
WATER_PRESSURE_PROPERTY = "Water pressure"
WATER_TEMPERATURE_PROPERTY = "Water temperature"
WEIGHT_PROPERTY = "Weight"


class MultilevelSensorType(IntEnum):
    """Enum for known multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    ACCELERATION_X_AXIS = 52
    ACCELERATION_Y_AXIS = 53
    ACCELERATION_Z_AXIS = 54
    AIR_FLOW = 18
    AIR_TEMPERATURE = 1
    AMMONIA = 84
    ANGLE_POSITION = 21
    APPLIED_FORCE_ON_THE_SENSOR = 71
    ATMOSPHERIC_PRESSURE = 8
    BAROMETRIC_PRESSURE = 9
    BASIS_METABOLIC_RATE = 50
    BLOOD_PRESSURE = 45
    BODY_MASS_INDEX = 51
    BOILER_WATER_TEMPERATURE = 62
    BONE_MASS = 48
    CARBON_DIOXIDE_LEVEL = 17
    CARBON_MONOXIDE_LEVEL = 40
    CONDENSER_COIL_TEMPERATURE = 74
    CURRENT = 16
    DEFROST_TEMPERATURE = 80
    DEW_POINT = 11
    DIRECTION = 7
    DISCHARGE_LINE_TEMPERATURE = 77
    DISCHARGE_PRESSURE = 79
    DISTANCE = 20
    DOMESTIC_HOT_WATER_TEMPERATURE = 63
    ELECTRICAL_CONDUCTIVITY = 29
    ELECTRICAL_RESISTIVITY = 28
    EVAPORATOR_COIL_TEMPERATURE = 75
    EXHAUST_TEMPERATURE = 65
    FAT_MASS = 47
    FORMALDEHYDE_LEVEL = 36
    FREQUENCY = 32
    GENERAL_PURPOSE = 2
    HEART_RATE = 44
    HEART_RATE_LF_HF_RATIO = 69
    HUMIDITY = 5
    ILLUMINANCE = 3
    LEAD = 85
    LIQUID_LINE_TEMPERATURE = 76
    LOUDNESS = 30
    METHANE_DENSITY = 38
    MOISTURE = 31
    MOTION_DIRECTION = 70
    MUSCLE_MASS = 46
    NITROGEN_DIOXIDE = 83
    OUTSIDE_TEMPERATURE = 64
    OZONE = 81
    PARTICULATE_MATTER_1 = 86
    PARTICULATE_MATTER_10 = 59
    PARTICULATE_MATTER_2_5 = 35
    PERSON_COUNTER_ENTERING = 87
    PERSON_COUNTER_EXITING = 88
    POWER = 4
    RADON_CONCENTRATION = 37
    RAIN_RATE = 12
    RELATIVE_MODULATION_LEVEL = 61
    RESPIRATORY_RATE = 60
    RETURN_AIR_TEMPERATURE = 72
    RF_SIGNAL_STRENGTH = 58
    ROTATION = 22
    SEISMIC_INTENSITY = 25
    SEISMIC_MAGNITUDE = 26
    SMOKE_DENSITY = 55
    SOIL_HUMIDITY = 41
    SOIL_REACTIVITY = 42
    SOIL_SALINITY = 43
    SOIL_TEMPERATURE = 24
    SOLAR_RADIATION = 10
    SUCTION_PRESSURE = 78
    SULFUR_DIOXIDE = 82
    SUPPLY_AIR_TEMPERATURE = 73
    TANK_CAPACITY = 19
    TARGET_TEMPERATURE = 34
    TIDE_LEVEL = 13
    TIME = 33
    TOTAL_BODY_WATER = 49
    ULTRAVIOLET = 27
    VELOCITY = 6
    VOLATILE_ORGANIC_COMPOUND_LEVEL = 39
    VOLTAGE = 15
    WATER_ACIDITY = 67
    WATER_CHLORINE_LEVEL = 66
    WATER_FLOW = 56
    WATER_OXIDATION_REDUCTION_POTENTIAL = 68
    WATER_PRESSURE = 57
    WATER_TEMPERATURE = 23
    WEIGHT = 14


class MultilevelSensorScaleType(IntEnum):
    """Common base class for multilevel sensor scale enums."""


class AccelerationScale(MultilevelSensorScaleType):
    """Enum for known scales for ACCELERATION multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    METER_PER_SQUARE_SECOND = 0


class AcidityScale(MultilevelSensorScaleType):
    """Enum for known scales for ACIDITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    ACIDITY = 0


class AirFlowScale(MultilevelSensorScaleType):
    """Enum for known scales for AIR_FLOW multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    CUBIC_FEET_PER_MINUTE = 1
    CUBIC_METER_PER_HOUR = 0


class AirPressureScale(MultilevelSensorScaleType):
    """Enum for known scales for AIR_PRESSURE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    INCHES_OF_MERCURY = 1
    KILOPASCAL = 0


class AnglePositionScale(MultilevelSensorScaleType):
    """Enum for known scales for ANGLE_POSITION multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    DEGREES_RELATIVE_TO_NORTH_POLE_OF_STANDING_EYE_VIEW = 1
    DEGREES_RELATIVE_TO_SOUTH_POLE_OF_STANDING_EYE_VIEW = 2
    PERCENTAGE_VALUE = 0


class AppliedForceOnTheSensorScale(MultilevelSensorScaleType):
    """Enum for known scales for APPLIED_FORCE_ON_THE_SENSOR multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    NEWTON = 0


class BasisMetabolicRateScale(MultilevelSensorScaleType):
    """Enum for known scales for BASIS_METABOLIC_RATE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    JOULE = 0


class BloodPressureScale(MultilevelSensorScaleType):
    """Enum for known scales for BLOOD_PRESSURE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    DIASTOLIC = 1
    SYSTOLIC = 0


class BodyMassIndexScale(MultilevelSensorScaleType):
    """Enum for known scales for BODY_MASS_INDEX multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    BODY_MASS_INDEX = 0


class CarbonDioxideLevelScale(MultilevelSensorScaleType):
    """Enum for known scales for CARBON_DIOXIDE_LEVEL multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    PARTS_MILLION = 0


class CarbonMonoxideLevelScale(MultilevelSensorScaleType):
    """Enum for known scales for CARBON_MONOXIDE_LEVEL multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MOLE_PER_CUBIC_METER = 0
    PARTS_MILLION = 1


class CurrentScale(MultilevelSensorScaleType):
    """Enum for known scales for CURRENT multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    AMPERE = 0
    MILLIAMPERE = 1


class DensityScale(MultilevelSensorScaleType):
    """Enum for known scales for DENSITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    DENSITY = 0


class DirectionScale(MultilevelSensorScaleType):
    """Enum for known scales for DIRECTION multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    DEGREES = 0


class DistanceScale(MultilevelSensorScaleType):
    """Enum for known scales for DISTANCE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    CENTIMETER = 1
    FEET = 2
    METER = 0


class ElectricalConductivityScale(MultilevelSensorScaleType):
    """Enum for known scales for ELECTRICAL_CONDUCTIVITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    SIEMENS_PER_METER = 0


class ElectricalResistivityScale(MultilevelSensorScaleType):
    """Enum for known scales for ELECTRICAL_RESISTIVITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    OHM_METER = 0


class FormaldehydeLevelScale(MultilevelSensorScaleType):
    """Enum for known scales for FORMALDEHYDE_LEVEL multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MOLE_PER_CUBIC_METER = 0


class FrequencyScale(MultilevelSensorScaleType):
    """Enum for known scales for FREQUENCY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    HERTZ = 0
    KILOHERTZ = 1


class GeneralPurposeScale(MultilevelSensorScaleType):
    """Enum for known scales for GENERAL_PURPOSE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    DIMENSIONLESS_VALUE = 1
    PERCENTAGE_VALUE = 0


class HeartRateScale(MultilevelSensorScaleType):
    """Enum for known scales for HEART_RATE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    BEATS_PER_MINUTE = 0


class HumidityScale(MultilevelSensorScaleType):
    """Enum for known scales for HUMIDITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    ABSOLUTE_HUMIDITY = 1
    PERCENTAGE_VALUE = 0


class IlluminanceScale(MultilevelSensorScaleType):
    """Enum for known scales for ILLUMINANCE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    LUX = 1
    PERCENTAGE_VALUE = 0


class LoudnessScale(MultilevelSensorScaleType):
    """Enum for known scales for LOUDNESS multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    A_WEIGHTED_DECIBELS = 1
    DECIBEL = 0


class MassScale(MultilevelSensorScaleType):
    """Enum for known scales for MASS multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    KILOGRAM = 0


class MethaneDensityScale(MultilevelSensorScaleType):
    """Enum for known scales for METHANE_DENSITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MOLE_PER_CUBIC_METER = 0


class MoistureScale(MultilevelSensorScaleType):
    """Enum for known scales for MOISTURE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    IMPEDANCE = 2
    PERCENTAGE_VALUE = 0
    VOLUME_WATER_CONTENT = 1
    WATER_ACTIVITY = 3


class ParticulateMatter10Scale(MultilevelSensorScaleType):
    """Enum for known scales for PARTICULATE_MATTER_10 multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MICROGRAM_PER_CUBIC_METER = 1
    MOLE_PER_CUBIC_METER = 0


class ParticulateMatter25Scale(MultilevelSensorScaleType):
    """Enum for known scales for PARTICULATE_MATTER_2_5 multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MICROGRAM_PER_CUBIC_METER = 1
    MOLE_PER_CUBIC_METER = 0


class PercentageScale(MultilevelSensorScaleType):
    """Enum for known scales for PERCENTAGE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    PERCENTAGE_VALUE = 0


class PowerScale(MultilevelSensorScaleType):
    """Enum for known scales for POWER multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    BTU_H = 1
    WATT = 0


class PressureScale(MultilevelSensorScaleType):
    """Enum for known scales for PRESSURE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    KILOPASCAL = 0
    POUND_PER_SQUARE_INCH = 1


class RadonConcentrationScale(MultilevelSensorScaleType):
    """Enum for known scales for RADON_CONCENTRATION multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    BECQUEREL_PER_CUBIC_METER = 0
    PICOCURIES_PER_LITER = 1


class RainRateScale(MultilevelSensorScaleType):
    """Enum for known scales for RAIN_RATE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    INCHES_PER_HOUR = 1
    MILLIMETER_HOUR = 0


class RespiratoryRateScale(MultilevelSensorScaleType):
    """Enum for known scales for RESPIRATORY_RATE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    BREATHS_PER_MINUTE = 0


class RfSignalStrengthScale(MultilevelSensorScaleType):
    """Enum for known scales for RF_SIGNAL_STRENGTH multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    POWER_LEVEL = 1
    RSSI = 0


class RotationScale(MultilevelSensorScaleType):
    """Enum for known scales for ROTATION multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    HERTZ = 1
    REVOLUTIONS_PER_MINUTE = 0


class SeismicIntensityScale(MultilevelSensorScaleType):
    """Enum for known scales for SEISMIC_INTENSITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    EUROPEAN_MACROSEISMIC = 1
    LIEDU = 2
    MERCALLI = 0
    SHINDO = 3


class SeismicMagnitudeScale(MultilevelSensorScaleType):
    """Enum for known scales for SEISMIC_MAGNITUDE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    BODY_WAVE = 3
    LOCAL = 0
    MOMENT = 1
    SURFACE_WAVE = 2


class SoilSalinityScale(MultilevelSensorScaleType):
    """Enum for known scales for SOIL_SALINITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MOLE_PER_CUBIC_METER = 0


class SolarRadiationScale(MultilevelSensorScaleType):
    """Enum for known scales for SOLAR_RADIATION multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    WATT_PER_SQUARE_METER = 0


class TankCapacityScale(MultilevelSensorScaleType):
    """Enum for known scales for TANK_CAPACITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    CUBIC_METER = 1
    GALLONS = 2
    LITER = 0


class TemperatureScale(MultilevelSensorScaleType):
    """Enum for known scales for TEMPERATURE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    CELSIUS = 0
    FAHRENHEIT = 1


class TideLevelScale(MultilevelSensorScaleType):
    """Enum for known scales for TIDE_LEVEL multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    FEET = 1
    METER = 0


class TimeScale(MultilevelSensorScaleType):
    """Enum for known scales for TIME multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    SECOND = 0


class UltravioletScale(MultilevelSensorScaleType):
    """Enum for known scales for ULTRAVIOLET multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    UV_INDEX = 0


class UnitlessScale(MultilevelSensorScaleType):
    """Enum for known scales for UNITLESS multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    UNITLESS = 0


class VelocityScale(MultilevelSensorScaleType):
    """Enum for known scales for VELOCITY multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MPH = 1
    M_S = 0


class VolatileOrganicCompoundLevelScale(MultilevelSensorScaleType):
    """Enum for known scales for VOLATILE_ORGANIC_COMPOUND_LEVEL multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MOLE_PER_CUBIC_METER = 0
    PARTS_MILLION = 1


class VoltageScale(MultilevelSensorScaleType):
    """Enum for known scales for VOLTAGE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MILLIVOLT = 1
    VOLT = 0


class WaterChlorineLevelScale(MultilevelSensorScaleType):
    """Enum for known scales for WATER_CHLORINE_LEVEL multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MILLIGRAM_PER_LITER = 0


class WaterFlowScale(MultilevelSensorScaleType):
    """Enum for known scales for WATER_FLOW multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    LITER_PER_HOUR = 0


class WaterOxidationReductionPotentialScale(MultilevelSensorScaleType):
    """Enum for known scales for WATER_OXIDATION_REDUCTION_POTENTIAL multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    MILLIVOLT = 0


class WaterPressureScale(MultilevelSensorScaleType):
    """Enum for known scales for WATER_PRESSURE multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    KILOPASCAL = 0


class WeightScale(MultilevelSensorScaleType):
    """Enum for known scales for WEIGHT multilevel sensor types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/registries/SensorTypes.ts
    KILOGRAM = 0
    POUNDS = 1


MULTILEVEL_SENSOR_TYPE_TO_SCALE_MAP: dict[
    MultilevelSensorType, type[MultilevelSensorScaleType]
] = {
    MultilevelSensorType.ACCELERATION_X_AXIS: AccelerationScale,
    MultilevelSensorType.ACCELERATION_Y_AXIS: AccelerationScale,
    MultilevelSensorType.ACCELERATION_Z_AXIS: AccelerationScale,
    MultilevelSensorType.AIR_FLOW: AirFlowScale,
    MultilevelSensorType.AIR_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.AMMONIA: DensityScale,
    MultilevelSensorType.ANGLE_POSITION: AnglePositionScale,
    MultilevelSensorType.APPLIED_FORCE_ON_THE_SENSOR: AppliedForceOnTheSensorScale,
    MultilevelSensorType.ATMOSPHERIC_PRESSURE: AirPressureScale,
    MultilevelSensorType.BAROMETRIC_PRESSURE: AirPressureScale,
    MultilevelSensorType.BASIS_METABOLIC_RATE: BasisMetabolicRateScale,
    MultilevelSensorType.BLOOD_PRESSURE: BloodPressureScale,
    MultilevelSensorType.BODY_MASS_INDEX: BodyMassIndexScale,
    MultilevelSensorType.BOILER_WATER_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.BONE_MASS: MassScale,
    MultilevelSensorType.CARBON_DIOXIDE_LEVEL: CarbonDioxideLevelScale,
    MultilevelSensorType.CARBON_MONOXIDE_LEVEL: CarbonMonoxideLevelScale,
    MultilevelSensorType.CONDENSER_COIL_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.CURRENT: CurrentScale,
    MultilevelSensorType.DEFROST_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.DEW_POINT: TemperatureScale,
    MultilevelSensorType.DIRECTION: DirectionScale,
    MultilevelSensorType.DISCHARGE_LINE_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.DISCHARGE_PRESSURE: PressureScale,
    MultilevelSensorType.DISTANCE: DistanceScale,
    MultilevelSensorType.DOMESTIC_HOT_WATER_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.ELECTRICAL_CONDUCTIVITY: ElectricalConductivityScale,
    MultilevelSensorType.ELECTRICAL_RESISTIVITY: ElectricalResistivityScale,
    MultilevelSensorType.EVAPORATOR_COIL_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.EXHAUST_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.FAT_MASS: MassScale,
    MultilevelSensorType.FORMALDEHYDE_LEVEL: FormaldehydeLevelScale,
    MultilevelSensorType.FREQUENCY: FrequencyScale,
    MultilevelSensorType.GENERAL_PURPOSE: GeneralPurposeScale,
    MultilevelSensorType.HEART_RATE: HeartRateScale,
    MultilevelSensorType.HEART_RATE_LF_HF_RATIO: UnitlessScale,
    MultilevelSensorType.HUMIDITY: HumidityScale,
    MultilevelSensorType.ILLUMINANCE: IlluminanceScale,
    MultilevelSensorType.LEAD: DensityScale,
    MultilevelSensorType.LIQUID_LINE_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.LOUDNESS: LoudnessScale,
    MultilevelSensorType.METHANE_DENSITY: MethaneDensityScale,
    MultilevelSensorType.MOISTURE: MoistureScale,
    MultilevelSensorType.MOTION_DIRECTION: DirectionScale,
    MultilevelSensorType.MUSCLE_MASS: MassScale,
    MultilevelSensorType.NITROGEN_DIOXIDE: DensityScale,
    MultilevelSensorType.OUTSIDE_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.OZONE: DensityScale,
    MultilevelSensorType.PARTICULATE_MATTER_1: DensityScale,
    MultilevelSensorType.PARTICULATE_MATTER_10: ParticulateMatter10Scale,
    MultilevelSensorType.PARTICULATE_MATTER_2_5: ParticulateMatter25Scale,
    MultilevelSensorType.PERSON_COUNTER_ENTERING: UnitlessScale,
    MultilevelSensorType.PERSON_COUNTER_EXITING: UnitlessScale,
    MultilevelSensorType.POWER: PowerScale,
    MultilevelSensorType.RADON_CONCENTRATION: RadonConcentrationScale,
    MultilevelSensorType.RAIN_RATE: RainRateScale,
    MultilevelSensorType.RELATIVE_MODULATION_LEVEL: PercentageScale,
    MultilevelSensorType.RESPIRATORY_RATE: RespiratoryRateScale,
    MultilevelSensorType.RETURN_AIR_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.RF_SIGNAL_STRENGTH: RfSignalStrengthScale,
    MultilevelSensorType.ROTATION: RotationScale,
    MultilevelSensorType.SEISMIC_INTENSITY: SeismicIntensityScale,
    MultilevelSensorType.SEISMIC_MAGNITUDE: SeismicMagnitudeScale,
    MultilevelSensorType.SMOKE_DENSITY: PercentageScale,
    MultilevelSensorType.SOIL_HUMIDITY: PercentageScale,
    MultilevelSensorType.SOIL_REACTIVITY: AcidityScale,
    MultilevelSensorType.SOIL_SALINITY: SoilSalinityScale,
    MultilevelSensorType.SOIL_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.SOLAR_RADIATION: SolarRadiationScale,
    MultilevelSensorType.SUCTION_PRESSURE: PressureScale,
    MultilevelSensorType.SULFUR_DIOXIDE: DensityScale,
    MultilevelSensorType.SUPPLY_AIR_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.TANK_CAPACITY: TankCapacityScale,
    MultilevelSensorType.TARGET_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.TIDE_LEVEL: TideLevelScale,
    MultilevelSensorType.TIME: TimeScale,
    MultilevelSensorType.TOTAL_BODY_WATER: MassScale,
    MultilevelSensorType.ULTRAVIOLET: UltravioletScale,
    MultilevelSensorType.VELOCITY: VelocityScale,
    MultilevelSensorType.VOLATILE_ORGANIC_COMPOUND_LEVEL: VolatileOrganicCompoundLevelScale,
    MultilevelSensorType.VOLTAGE: VoltageScale,
    MultilevelSensorType.WATER_ACIDITY: AcidityScale,
    MultilevelSensorType.WATER_CHLORINE_LEVEL: WaterChlorineLevelScale,
    MultilevelSensorType.WATER_FLOW: WaterFlowScale,
    MultilevelSensorType.WATER_OXIDATION_REDUCTION_POTENTIAL: WaterOxidationReductionPotentialScale,
    MultilevelSensorType.WATER_PRESSURE: WaterPressureScale,
    MultilevelSensorType.WATER_TEMPERATURE: TemperatureScale,
    MultilevelSensorType.WEIGHT: WeightScale,
}

UNIT_ABSOLUTE_HUMIDITY: list[MultilevelSensorScaleType] = [
    HumidityScale.ABSOLUTE_HUMIDITY
]
UNIT_ACIDITY: list[MultilevelSensorScaleType] = [AcidityScale.ACIDITY]
UNIT_AMPERE: list[MultilevelSensorScaleType] = [CurrentScale.AMPERE]
UNIT_A_WEIGHTED_DECIBELS: list[MultilevelSensorScaleType] = [
    LoudnessScale.A_WEIGHTED_DECIBELS
]
UNIT_BEATS_PER_MINUTE: list[MultilevelSensorScaleType] = [
    HeartRateScale.BEATS_PER_MINUTE
]
UNIT_BECQUEREL_PER_CUBIC_METER: list[MultilevelSensorScaleType] = [
    RadonConcentrationScale.BECQUEREL_PER_CUBIC_METER
]
UNIT_BODY_MASS_INDEX: list[MultilevelSensorScaleType] = [
    BodyMassIndexScale.BODY_MASS_INDEX
]
UNIT_BODY_WAVE: list[MultilevelSensorScaleType] = [SeismicMagnitudeScale.BODY_WAVE]
UNIT_BREATHS_PER_MINUTE: list[MultilevelSensorScaleType] = [
    RespiratoryRateScale.BREATHS_PER_MINUTE
]
UNIT_BTU_H: list[MultilevelSensorScaleType] = [PowerScale.BTU_H]
UNIT_CELSIUS: list[MultilevelSensorScaleType] = [TemperatureScale.CELSIUS]
UNIT_CENTIMETER: list[MultilevelSensorScaleType] = [DistanceScale.CENTIMETER]
UNIT_CUBIC_FEET_PER_MINUTE: list[MultilevelSensorScaleType] = [
    AirFlowScale.CUBIC_FEET_PER_MINUTE
]
UNIT_CUBIC_METER: list[MultilevelSensorScaleType] = [TankCapacityScale.CUBIC_METER]
UNIT_CUBIC_METER_PER_HOUR: list[MultilevelSensorScaleType] = [
    AirFlowScale.CUBIC_METER_PER_HOUR
]
UNIT_DECIBEL: list[MultilevelSensorScaleType] = [LoudnessScale.DECIBEL]
UNIT_DEGREES: list[MultilevelSensorScaleType] = [DirectionScale.DEGREES]
UNIT_DEGREES_RELATIVE_TO_NORTH_POLE_OF_STANDING_EYE_VIEW: list[
    MultilevelSensorScaleType
] = [AnglePositionScale.DEGREES_RELATIVE_TO_NORTH_POLE_OF_STANDING_EYE_VIEW]
UNIT_DEGREES_RELATIVE_TO_SOUTH_POLE_OF_STANDING_EYE_VIEW: list[
    MultilevelSensorScaleType
] = [AnglePositionScale.DEGREES_RELATIVE_TO_SOUTH_POLE_OF_STANDING_EYE_VIEW]
UNIT_DENSITY: list[MultilevelSensorScaleType] = [DensityScale.DENSITY]
UNIT_DIASTOLIC: list[MultilevelSensorScaleType] = [BloodPressureScale.DIASTOLIC]
UNIT_DIMENSIONLESS_VALUE: list[MultilevelSensorScaleType] = [
    GeneralPurposeScale.DIMENSIONLESS_VALUE
]
UNIT_EUROPEAN_MACROSEISMIC: list[MultilevelSensorScaleType] = [
    SeismicIntensityScale.EUROPEAN_MACROSEISMIC
]
UNIT_FAHRENHEIT: list[MultilevelSensorScaleType] = [TemperatureScale.FAHRENHEIT]
UNIT_FEET: list[MultilevelSensorScaleType] = [DistanceScale.FEET, TideLevelScale.FEET]
UNIT_GALLONS: list[MultilevelSensorScaleType] = [TankCapacityScale.GALLONS]
UNIT_HERTZ: list[MultilevelSensorScaleType] = [
    FrequencyScale.HERTZ,
    RotationScale.HERTZ,
]
UNIT_IMPEDANCE: list[MultilevelSensorScaleType] = [MoistureScale.IMPEDANCE]
UNIT_INCHES_OF_MERCURY: list[MultilevelSensorScaleType] = [
    AirPressureScale.INCHES_OF_MERCURY
]
UNIT_INCHES_PER_HOUR: list[MultilevelSensorScaleType] = [RainRateScale.INCHES_PER_HOUR]
UNIT_JOULE: list[MultilevelSensorScaleType] = [BasisMetabolicRateScale.JOULE]
UNIT_KILOGRAM: list[MultilevelSensorScaleType] = [
    MassScale.KILOGRAM,
    WeightScale.KILOGRAM,
]
UNIT_KILOHERTZ: list[MultilevelSensorScaleType] = [FrequencyScale.KILOHERTZ]
UNIT_KILOPASCAL: list[MultilevelSensorScaleType] = [
    AirPressureScale.KILOPASCAL,
    PressureScale.KILOPASCAL,
    WaterPressureScale.KILOPASCAL,
]
UNIT_LIEDU: list[MultilevelSensorScaleType] = [SeismicIntensityScale.LIEDU]
UNIT_LITER: list[MultilevelSensorScaleType] = [TankCapacityScale.LITER]
UNIT_LITER_PER_HOUR: list[MultilevelSensorScaleType] = [WaterFlowScale.LITER_PER_HOUR]
UNIT_LOCAL: list[MultilevelSensorScaleType] = [SeismicMagnitudeScale.LOCAL]
UNIT_LUX: list[MultilevelSensorScaleType] = [IlluminanceScale.LUX]
UNIT_MERCALLI: list[MultilevelSensorScaleType] = [SeismicIntensityScale.MERCALLI]
UNIT_METER: list[MultilevelSensorScaleType] = [
    DistanceScale.METER,
    TideLevelScale.METER,
]
UNIT_METER_PER_SQUARE_SECOND: list[MultilevelSensorScaleType] = [
    AccelerationScale.METER_PER_SQUARE_SECOND
]
UNIT_MICROGRAM_PER_CUBIC_METER: list[MultilevelSensorScaleType] = [
    ParticulateMatter10Scale.MICROGRAM_PER_CUBIC_METER,
    ParticulateMatter25Scale.MICROGRAM_PER_CUBIC_METER,
]
UNIT_MILLIAMPERE: list[MultilevelSensorScaleType] = [CurrentScale.MILLIAMPERE]
UNIT_MILLIGRAM_PER_LITER: list[MultilevelSensorScaleType] = [
    WaterChlorineLevelScale.MILLIGRAM_PER_LITER
]
UNIT_MILLIMETER_HOUR: list[MultilevelSensorScaleType] = [RainRateScale.MILLIMETER_HOUR]
UNIT_MILLIVOLT: list[MultilevelSensorScaleType] = [
    VoltageScale.MILLIVOLT,
    WaterOxidationReductionPotentialScale.MILLIVOLT,
]
UNIT_MOLE_PER_CUBIC_METER: list[MultilevelSensorScaleType] = [
    CarbonMonoxideLevelScale.MOLE_PER_CUBIC_METER,
    FormaldehydeLevelScale.MOLE_PER_CUBIC_METER,
    MethaneDensityScale.MOLE_PER_CUBIC_METER,
    ParticulateMatter10Scale.MOLE_PER_CUBIC_METER,
    ParticulateMatter25Scale.MOLE_PER_CUBIC_METER,
    SoilSalinityScale.MOLE_PER_CUBIC_METER,
    VolatileOrganicCompoundLevelScale.MOLE_PER_CUBIC_METER,
]
UNIT_MOMENT: list[MultilevelSensorScaleType] = [SeismicMagnitudeScale.MOMENT]
UNIT_MPH: list[MultilevelSensorScaleType] = [VelocityScale.MPH]
UNIT_M_S: list[MultilevelSensorScaleType] = [VelocityScale.M_S]
UNIT_NEWTON: list[MultilevelSensorScaleType] = [AppliedForceOnTheSensorScale.NEWTON]
UNIT_OHM_METER: list[MultilevelSensorScaleType] = [ElectricalResistivityScale.OHM_METER]
UNIT_PARTS_MILLION: list[MultilevelSensorScaleType] = [
    CarbonDioxideLevelScale.PARTS_MILLION,
    CarbonMonoxideLevelScale.PARTS_MILLION,
    VolatileOrganicCompoundLevelScale.PARTS_MILLION,
]
UNIT_PERCENTAGE_VALUE: list[MultilevelSensorScaleType] = [
    AnglePositionScale.PERCENTAGE_VALUE,
    GeneralPurposeScale.PERCENTAGE_VALUE,
    HumidityScale.PERCENTAGE_VALUE,
    IlluminanceScale.PERCENTAGE_VALUE,
    MoistureScale.PERCENTAGE_VALUE,
    PercentageScale.PERCENTAGE_VALUE,
]
UNIT_PICOCURIES_PER_LITER: list[MultilevelSensorScaleType] = [
    RadonConcentrationScale.PICOCURIES_PER_LITER
]
UNIT_POUNDS: list[MultilevelSensorScaleType] = [WeightScale.POUNDS]
UNIT_POUND_PER_SQUARE_INCH: list[MultilevelSensorScaleType] = [
    PressureScale.POUND_PER_SQUARE_INCH
]
UNIT_POWER_LEVEL: list[MultilevelSensorScaleType] = [RfSignalStrengthScale.POWER_LEVEL]
UNIT_REVOLUTIONS_PER_MINUTE: list[MultilevelSensorScaleType] = [
    RotationScale.REVOLUTIONS_PER_MINUTE
]
UNIT_RSSI: list[MultilevelSensorScaleType] = [RfSignalStrengthScale.RSSI]
UNIT_SECOND: list[MultilevelSensorScaleType] = [TimeScale.SECOND]
UNIT_SHINDO: list[MultilevelSensorScaleType] = [SeismicIntensityScale.SHINDO]
UNIT_SIEMENS_PER_METER: list[MultilevelSensorScaleType] = [
    ElectricalConductivityScale.SIEMENS_PER_METER
]
UNIT_SURFACE_WAVE: list[MultilevelSensorScaleType] = [
    SeismicMagnitudeScale.SURFACE_WAVE
]
UNIT_SYSTOLIC: list[MultilevelSensorScaleType] = [BloodPressureScale.SYSTOLIC]
UNIT_UNITLESS: list[MultilevelSensorScaleType] = [UnitlessScale.UNITLESS]
UNIT_UV_INDEX: list[MultilevelSensorScaleType] = [UltravioletScale.UV_INDEX]
UNIT_VOLT: list[MultilevelSensorScaleType] = [VoltageScale.VOLT]
UNIT_VOLUME_WATER_CONTENT: list[MultilevelSensorScaleType] = [
    MoistureScale.VOLUME_WATER_CONTENT
]
UNIT_WATER_ACTIVITY: list[MultilevelSensorScaleType] = [MoistureScale.WATER_ACTIVITY]
UNIT_WATT: list[MultilevelSensorScaleType] = [PowerScale.WATT]
UNIT_WATT_PER_SQUARE_METER: list[MultilevelSensorScaleType] = [
    SolarRadiationScale.WATT_PER_SQUARE_METER
]

# ----------------------------------------------------------------------------------- #
# **END OF AUTOGENERATED CONTENT** (DO NOT EDIT/REMOVE THIS COMMENT BLOCK AND DO NOT  #
# EDIT ANYTHING ABOVE IT. IF A NEW IMPORT IS NEEDED, ADD IT TO THE IMPORTS IN THE     #
# CORRESPONDING GENERATION SCRIPT THEN RE-RUN THE SCRIPT. LINES WRITTEN BELOW THIS    #
# BLOCK WILL BE PRESERVED AS LONG AS THIS BLOCK REMAINS)                              #
# ----------------------------------------------------------------------------------- #

CO_SENSORS = [MultilevelSensorType.CARBON_MONOXIDE_LEVEL]
CO2_SENSORS = [MultilevelSensorType.CARBON_DIOXIDE_LEVEL]
CURRENT_SENSORS = [MultilevelSensorType.CURRENT]
ENERGY_MEASUREMENT_SENSORS = [MultilevelSensorType.BASIS_METABOLIC_RATE]
HUMIDITY_SENSORS = [MultilevelSensorType.HUMIDITY]
ILLUMINANCE_SENSORS = [MultilevelSensorType.ILLUMINANCE]
POWER_SENSORS = [MultilevelSensorType.POWER]
PRESSURE_SENSORS = [
    MultilevelSensorType.BLOOD_PRESSURE,
    MultilevelSensorType.WATER_PRESSURE,
    MultilevelSensorType.SUCTION_PRESSURE,
    MultilevelSensorType.DISCHARGE_PRESSURE,
    MultilevelSensorType.BAROMETRIC_PRESSURE,
    MultilevelSensorType.ATMOSPHERIC_PRESSURE,
]
SIGNAL_STRENGTH_SENSORS = [MultilevelSensorType.RF_SIGNAL_STRENGTH]
TEMPERATURE_SENSORS = [
    MultilevelSensorType.AIR_TEMPERATURE,
    MultilevelSensorType.DEW_POINT,
    MultilevelSensorType.WATER_TEMPERATURE,
    MultilevelSensorType.SOIL_TEMPERATURE,
    MultilevelSensorType.TARGET_TEMPERATURE,
    MultilevelSensorType.BOILER_WATER_TEMPERATURE,
    MultilevelSensorType.DOMESTIC_HOT_WATER_TEMPERATURE,
    MultilevelSensorType.OUTSIDE_TEMPERATURE,
    MultilevelSensorType.EXHAUST_TEMPERATURE,
    MultilevelSensorType.RETURN_AIR_TEMPERATURE,
    MultilevelSensorType.SUPPLY_AIR_TEMPERATURE,
    MultilevelSensorType.CONDENSER_COIL_TEMPERATURE,
    MultilevelSensorType.EVAPORATOR_COIL_TEMPERATURE,
    MultilevelSensorType.LIQUID_LINE_TEMPERATURE,
    MultilevelSensorType.DISCHARGE_LINE_TEMPERATURE,
    MultilevelSensorType.DEFROST_TEMPERATURE,
]
VOLTAGE_SENSORS = [
    MultilevelSensorType.VOLTAGE,
    MultilevelSensorType.WATER_OXIDATION_REDUCTION_POTENTIAL,
]



"""Provide a model for the Z-Wave JS Driver."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, cast

from zwave_js_server.model.firmware import (
    FirmwareUpdateData,
    FirmwareUpdateDataDataType,
    FirmwareUpdateInfo,
    FirmwareUpdateInfoDataType,
)

from ...event import BaseEventModel, Event, EventBase
from ..config_manager import ConfigManager
from ..controller import Controller
from ..log_config import LogConfig, LogConfigDataType
from ..log_message import LogMessage, LogMessageContextDataType, LogMessageDataType
from .firmware import (
    DriverFirmwareUpdateProgress,
    DriverFirmwareUpdateProgressDataType,
    DriverFirmwareUpdateResult,
    DriverFirmwareUpdateResultDataType,
)

if TYPE_CHECKING:
    from ...client import Client


class BaseDriverEventModel(BaseEventModel):
    """Base model for a driver event."""

    source: Literal["driver"]


class LogConfigUpdatedEventModel(BaseDriverEventModel):
    """Model for `log config updated` event data."""

    event: Literal["log config updated"]
    config: LogConfigDataType

    @classmethod
    def from_dict(cls, data: dict) -> LogConfigUpdatedEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            config=data["config"],
        )


class AllNodesReadyEventModel(BaseDriverEventModel):
    """Model for `all nodes ready` event data."""

    event: Literal["all nodes ready"]


class LoggingEventModel(BaseDriverEventModel):
    """Model for `logging` event data."""

    event: Literal["logging"]
    message: str | list[str]  # required
    formattedMessage: str | list[str]  # required
    direction: str  # required
    level: str  # required
    context: LogMessageContextDataType  # required
    primaryTags: str | None = None
    secondaryTags: str | None = None
    secondaryTagPadding: int | None = None
    multiline: bool | None = None
    timestamp: str | None = None
    label: str | None = None

    @classmethod
    def from_dict(cls, data: dict) -> LoggingEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            message=data["message"],
            formattedMessage=data["formattedMessage"],
            direction=data["direction"],
            level=data["level"],
            context=data["context"],
            primaryTags=data.get("primaryTags"),
            secondaryTags=data.get("secondaryTags"),
            secondaryTagPadding=data.get("secondaryTagPadding"),
            multiline=data.get("multiline"),
            timestamp=data.get("timestamp"),
            label=data.get("label"),
        )


class DriverReadyEventModel(BaseDriverEventModel):
    """Model for `driver ready` event data."""

    event: Literal["driver ready"]


class FirmwareUpdateFinishedEventModel(BaseDriverEventModel):
    """Model for `firmware update finished` event data."""

    event: Literal["firmware update finished"]
    result: DriverFirmwareUpdateResultDataType

    @classmethod
    def from_dict(cls, data: dict) -> FirmwareUpdateFinishedEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            result=data["result"],
        )


class FirmwareUpdateProgressEventModel(BaseDriverEventModel):
    """Model for `firmware update progress` event data."""

    event: Literal["firmware update progress"]
    progress: DriverFirmwareUpdateProgressDataType

    @classmethod
    def from_dict(cls, data: dict) -> FirmwareUpdateProgressEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            progress=data["progress"],
        )


DRIVER_EVENT_MODEL_MAP: dict[str, type[BaseDriverEventModel]] = {
    "all nodes ready": AllNodesReadyEventModel,
    "log config updated": LogConfigUpdatedEventModel,
    "logging": LoggingEventModel,
    "driver ready": DriverReadyEventModel,
    "firmware update finished": FirmwareUpdateFinishedEventModel,
    "firmware update progress": FirmwareUpdateProgressEventModel,
}


class CheckConfigUpdates:
    """Represent config updates check."""

    def __init__(self, data: dict) -> None:
        """Initialize class."""
        self.installed_version: str = data["installedVersion"]
        self.update_available: bool = data["updateAvailable"]
        self.new_version: str | None = data.get("newVersion")


class Driver(EventBase):
    """Represent a Z-Wave JS driver."""

    def __init__(
        self, client: Client, state: dict, log_config: LogConfigDataType
    ) -> None:
        """Initialize driver."""
        super().__init__()
        self.client = client
        self.controller = Controller(client, state)
        self.log_config = LogConfig.from_dict(log_config)
        self.config_manager = ConfigManager(client)
        self._firmware_update_progress: DriverFirmwareUpdateProgress | None = None

    def __hash__(self) -> int:
        """Return the hash."""
        return hash(self.controller)

    def __eq__(self, other: object) -> bool:
        """Return whether this instance equals another."""
        if not isinstance(other, Driver):
            return False
        return self.controller == other.controller

    @property
    def firmware_update_progress(self) -> DriverFirmwareUpdateProgress | None:
        """Return firmware update progress."""
        return self._firmware_update_progress

    def receive_event(self, event: Event) -> None:
        """Receive an event."""
        if event.data["source"] != "driver":
            self.controller.receive_event(event)
            return

        DRIVER_EVENT_MODEL_MAP[event.type].from_dict(event.data)

        self._handle_event_protocol(event)

        self.emit(event.type, event.data)

    async def _async_send_command(
        self, command: str, require_schema: int | None = None, **kwargs: Any
    ) -> dict:
        """Send a driver command. For internal use only."""
        return await self.client.async_send_command(
            {
                "command": f"driver.{command}",
                **kwargs,
            },
            require_schema,
        )

    async def async_update_log_config(self, log_config: LogConfig) -> None:
        """Update log config for driver."""
        await self._async_send_command(
            "update_log_config", config=log_config.to_dict(), require_schema=4
        )

    async def async_get_log_config(self) -> LogConfig:
        """Return current log config for driver."""
        result = await self._async_send_command("get_log_config", require_schema=4)
        return LogConfig.from_dict(result["config"])

    async def async_enable_statistics(
        self, application_name: str, application_version: str
    ) -> None:
        """Send command to enable data collection."""
        await self._async_send_command(
            "enable_statistics",
            applicationName=application_name,
            applicationVersion=application_version,
            require_schema=4,
        )

    async def async_disable_statistics(self) -> None:
        """Send command to stop listening to log events."""
        await self._async_send_command("disable_statistics", require_schema=4)

    async def async_is_statistics_enabled(self) -> bool:
        """Send command to start listening to log events."""
        result = await self._async_send_command(
            "is_statistics_enabled", require_schema=4
        )
        return cast(bool, result["statisticsEnabled"])

    async def async_check_for_config_updates(self) -> CheckConfigUpdates:
        """Send command to check for config updates."""
        result = await self._async_send_command(
            "check_for_config_updates", require_schema=5
        )
        return CheckConfigUpdates(result)

    async def async_install_config_update(self) -> bool:
        """Send command to install config update."""
        result = await self._async_send_command(
            "install_config_update", require_schema=5
        )
        return cast(bool, result["success"])

    async def async_firmware_update_otw(
        self,
        *,
        update_data: FirmwareUpdateData | None = None,
        update_info: FirmwareUpdateInfo | None = None,
    ) -> DriverFirmwareUpdateResult:
        """Send firmwareUpdateOTW command to Driver."""
        if update_data is None and update_info is None:
            raise ValueError(
                "Either update_data or update_info must be provided for firmware update."
            )
        if update_data is not None and update_info is not None:
            raise ValueError(
                "Only one of update_data or update_info can be provided for firmware update."
            )
        params: FirmwareUpdateDataDataType | dict[str, FirmwareUpdateInfoDataType]
        if update_data is not None:
            params = update_data.to_dict()
        elif update_info is not None:
            params = {"updateInfo": update_info.to_dict()}
        data = await self._async_send_command(
            "firmware_update_otw", require_schema=44, **params
        )
        return DriverFirmwareUpdateResult(data["result"])

    async def async_is_otw_firmware_update_in_progress(self) -> bool:
        """Send isOTWFirmwareUpdateInProgress command to Driver."""
        result = await self._async_send_command(
            "is_otw_firmware_update_in_progress", require_schema=41
        )
        return cast(bool, result["progress"])

    async def async_set_preferred_scales(
        self, scales: dict[str | int, str | int]
    ) -> None:
        """Send command to set preferred sensor scales."""
        await self._async_send_command(
            "set_preferred_scales", scales=scales, require_schema=6
        )

    async def async_hard_reset(self) -> None:
        """Send command to hard reset controller."""
        await self._async_send_command("hard_reset", require_schema=25)

    async def async_try_soft_reset(self) -> None:
        """Send command to try to soft reset controller."""
        await self._async_send_command("try_soft_reset", require_schema=25)

    async def async_soft_reset(self) -> None:
        """Send command to soft reset controller."""
        await self._async_send_command("soft_reset", require_schema=25)

    async def async_shutdown(self) -> bool:
        """Send command to shutdown controller."""
        data = await self._async_send_command("shutdown", require_schema=27)
        return cast(bool, data["success"])

    def handle_logging(self, event: Event) -> None:
        """Process a driver logging event."""
        event.data["log_message"] = LogMessage(cast(LogMessageDataType, event.data))

    def handle_log_config_updated(self, event: Event) -> None:
        """Process a driver log config updated event."""
        event.data["log_config"] = self.log_config = LogConfig.from_dict(
            event.data["config"]
        )

    def handle_all_nodes_ready(self, event: Event) -> None:
        """Process a driver all nodes ready event."""

    def handle_driver_ready(self, event: Event) -> None:
        """Process a driver ready event."""

    def handle_firmware_update_progress(self, event: Event) -> None:
        """Process a firmware update progress event."""
        self._firmware_update_progress = event.data["firmware_update_progress"] = (
            DriverFirmwareUpdateProgress(event.data["progress"])
        )

    def handle_firmware_update_finished(self, event: Event) -> None:
        """Process a firmware update finished event."""
        self._firmware_update_progress = None
        event.data["firmware_update_finished"] = DriverFirmwareUpdateResult(
            event.data["result"]
        )



"""Dump helper."""

from __future__ import annotations

import asyncio

import aiohttp

from .client import INITIALIZE_MESSAGE_ID
from .const import MAX_SERVER_SCHEMA_VERSION, PACKAGE_NAME, __version__


async def dump_msgs(
    url: str,
    session: aiohttp.ClientSession,
    additional_user_agent_components: dict[str, str] | None = None,
    timeout: float | None = None,
) -> list[dict]:
    """Dump server state."""
    client = await session.ws_connect(url, compress=15, max_msg_size=0)
    msgs = []

    version = await client.receive_json()
    msgs.append(version)

    for to_send in (
        {
            "command": "initialize",
            "messageId": INITIALIZE_MESSAGE_ID,
            "schemaVersion": MAX_SERVER_SCHEMA_VERSION,
            "additionalUserAgentComponents": {
                PACKAGE_NAME: __version__,
                **(additional_user_agent_components or {}),
            },
        },
        {"command": "start_listening", "messageId": "start-listening"},
    ):
        await client.send_json(to_send)
        msgs.append(await client.receive_json())

    if timeout is None:
        await client.close()
        return msgs

    current_task = asyncio.current_task()
    assert current_task is not None
    asyncio.get_running_loop().call_later(timeout, current_task.cancel)

    while True:
        try:
            msg = await client.receive_json()
            msgs.append(msg)
        except asyncio.CancelledError:
            break

    await client.close()
    return msgs



"""Provide a model for Z-Wave driver firmware."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum

from zwave_js_server.model.firmware import (
    FirmwareUpdateData,
    FirmwareUpdateDataDataType,
    FirmwareUpdateProgress,
    FirmwareUpdateProgressDataType,
    FirmwareUpdateResult,
    FirmwareUpdateResultDataType,
)


class DriverFirmwareUpdateDataDataType(FirmwareUpdateDataDataType):
    """Represent a driver firmware update data dict type."""


@dataclass
class DriverFirmwareUpdateData(FirmwareUpdateData):
    """Driver firmware update data."""


class DriverFirmwareUpdateStatus(IntEnum):
    """Enum with all driver firmware update status values.

    https://zwave-js.github.io/node-zwave-js/#/api/driver?id=quotfirmware-update-finishedquot
    """

    ERROR_TIMEOUT = 0
    # The maximum number of retry attempts for a firmware fragments were reached
    ERROR_RETRY_LIMIT_REACHED = 1
    # The update was aborted by the bootloader
    ERROR_ABORTED = 2
    # This driver does not support firmware updates
    ERROR_NOT_SUPPORTED = 3
    OK = 255


class DriverFirmwareUpdateProgressDataType(FirmwareUpdateProgressDataType):
    """Represent a driver firmware update progress dict type."""


@dataclass
class DriverFirmwareUpdateProgress(FirmwareUpdateProgress):
    """Model for a driver firmware update progress data."""

    data: DriverFirmwareUpdateProgressDataType = field(repr=False)


class DriverFirmwareUpdateResultDataType(FirmwareUpdateResultDataType):
    """Represent a driver firmware update result dict type."""


@dataclass
class DriverFirmwareUpdateResult(FirmwareUpdateResult):
    """Model for driver firmware update result data."""

    data: DriverFirmwareUpdateResultDataType = field(repr=False)
    status: DriverFirmwareUpdateStatus = field(init=False)
    success: bool = field(init=False)
    _status_class = DriverFirmwareUpdateStatus



"""
Model for Z-Wave JS config manager.

https://zwave-js.github.io/node-zwave-js/#/api/config-manager
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ..device_config import DeviceConfig

if TYPE_CHECKING:
    from ...client import Client


class ConfigManager:
    """Model for the Z-Wave JS config manager."""

    def __init__(self, client: Client) -> None:
        """Initialize."""
        self._client = client

    async def lookup_device(
        self,
        manufacturer_id: int,
        product_type: int,
        product_id: int,
        firmware_version: str | None = None,
    ) -> DeviceConfig | None:
        """Look up the definition of a given device in the configuration DB."""
        cmd: dict[str, Any] = {
            "command": "config_manager.lookup_device",
            "manufacturerId": manufacturer_id,
            "productType": product_type,
            "productId": product_id,
        }

        if firmware_version is not None:
            cmd["firmwareVersion"] = firmware_version

        data = await self._client.async_send_command(cmd)

        if not data or not (config := data.get("config")):
            return None

        return DeviceConfig(config)



"""Constants for the Power Level Command Class."""

from __future__ import annotations

from enum import IntEnum


class PowerLevelTestStatus(IntEnum):
    """Enum with all known power level test statuses."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/commandclass/PowerlevelCC.ts#L52
    FAILED = 0
    SUCCESS = 1
    IN_PROGRESS = 2



"""Energy Production Command Class specific utility functions for values."""

from __future__ import annotations

from ...const import CommandClass
from ...const.command_class.energy_production import (
    CC_SPECIFIC_PARAMETER,
    CC_SPECIFIC_SCALE,
    ENERGY_PRODUCTION_PARAMETER_TO_SCALE_ENUM_MAP,
    EnergyProductionParameter,
    EnergyProductionScaleType,
)
from ...exceptions import InvalidCommandClass, UnknownValueData
from ...model.value import Value


def get_energy_production_parameter(value: Value) -> EnergyProductionParameter:
    """Get the EnergyProductionParameter for a given value."""
    if value.command_class != CommandClass.ENERGY_PRODUCTION:
        raise InvalidCommandClass(value, CommandClass.ENERGY_PRODUCTION)
    try:
        return EnergyProductionParameter(
            value.metadata.cc_specific[CC_SPECIFIC_PARAMETER]
        )
    except ValueError:
        raise UnknownValueData(
            value, f"metadata.cc_specific.{CC_SPECIFIC_PARAMETER}"
        ) from None


def get_energy_production_scale_type(value: Value) -> EnergyProductionScaleType:
    """Get the ScaleType for a given value."""
    parameter = get_energy_production_parameter(value)
    scale_enum = ENERGY_PRODUCTION_PARAMETER_TO_SCALE_ENUM_MAP[parameter]
    try:
        return scale_enum(value.metadata.cc_specific[CC_SPECIFIC_SCALE])
    except ValueError:
        raise UnknownValueData(
            value, f"metadata.cc_specific.{CC_SPECIFIC_SCALE}"
        ) from None



"""Support for multicast commands."""

from __future__ import annotations

from typing import Any, cast

from ..client import Client
from ..const import CommandClass
from ..model.node import Node, _get_value_id_dict_from_value_data
from ..model.value import SetValueResult, ValueDataType


async def _async_send_command(
    client: Client,
    command: str,
    nodes: list[Node] | None = None,
    require_schema: int | None = None,
    **kwargs: Any,
) -> dict:
    """Send a multicast command."""
    if nodes:
        cmd = {
            "command": f"multicast_group.{command}",
            "nodeIDs": [node.node_id for node in nodes],
            **kwargs,
        }
    else:
        cmd = {"command": f"broadcast_node.{command}", **kwargs}

    return await client.async_send_command(cmd, require_schema)


async def async_multicast_set_value(
    client: Client,
    new_value: Any,
    value_data: ValueDataType,
    nodes: list[Node] | None = None,
    options: dict | None = None,
) -> SetValueResult:
    """Send a multicast set_value command."""
    result = await _async_send_command(
        client,
        "set_value",
        nodes,
        valueId=_get_value_id_dict_from_value_data(value_data),
        value=new_value,
        options=options,
        require_schema=29,
    )
    return SetValueResult(result["result"])


async def async_multicast_get_endpoint_count(
    client: Client, nodes: list[Node] | None = None
) -> int:
    """Send a multicast get_endpoint_count command."""
    result = await _async_send_command(
        client, "get_endpoint_count", nodes, require_schema=5
    )
    return cast(int, result["count"])


async def async_multicast_endpoint_supports_cc(
    client: Client,
    endpoint: int,
    command_class: CommandClass,
    nodes: list[Node] | None = None,
) -> bool:
    """Send a supports_cc command to a multicast endpoint."""
    result = await _async_send_command(
        client,
        "supports_cc",
        nodes,
        index=endpoint,
        commandClass=command_class,
        require_schema=5,
    )
    return cast(bool, result["supported"])


async def async_multicast_endpoint_get_cc_version(
    client: Client,
    endpoint: int,
    command_class: CommandClass,
    nodes: list[Node] | None = None,
) -> int:
    """Send a get_cc_version command to a multicast endpoint."""
    result = await _async_send_command(
        client,
        "get_cc_version",
        nodes,
        index=endpoint,
        commandClass=command_class,
        require_schema=5,
    )
    return cast(int, result["version"])


async def async_multicast_endpoint_invoke_cc_api(
    client: Client,
    endpoint: int,
    command_class: CommandClass,
    method_name: str,
    args: list[Any] | None = None,
    nodes: list[Node] | None = None,
) -> Any:
    """Send a invoke_cc_api command to a multicast endpoint."""
    result = await _async_send_command(
        client,
        "invoke_cc_api",
        nodes,
        index=endpoint,
        commandClass=command_class,
        methodName=method_name,
        args=args,
        require_schema=5,
    )
    return result["response"]


async def async_multicast_endpoint_supports_cc_api(
    client: Client,
    endpoint: int,
    command_class: CommandClass,
    nodes: list[Node] | None = None,
) -> bool:
    """Send a supports_cc_api command to a multicast endpoint."""
    result = await _async_send_command(
        client,
        "supports_cc_api",
        nodes,
        index=endpoint,
        commandClass=command_class,
        require_schema=5,
    )
    return cast(bool, result["supported"])



"""Utility functions for Z-Wave JS locks."""

from __future__ import annotations

from typing import TypedDict, cast

from ..const import CommandClass
from ..const.command_class.lock import (
    ATTR_CODE_SLOT,
    ATTR_IN_USE,
    ATTR_NAME,
    ATTR_USERCODE,
    CURRENT_AUTO_RELOCK_TIME_PROPERTY,
    CURRENT_BLOCK_TO_BLOCK_PROPERTY,
    CURRENT_HOLD_AND_RELEASE_TIME_PROPERTY,
    CURRENT_TWIST_ASSIST_PROPERTY,
    LOCK_USERCODE_ID_PROPERTY,
    LOCK_USERCODE_PROPERTY,
    LOCK_USERCODE_STATUS_PROPERTY,
    CodeSlotStatus,
    DoorLockCCConfigurationSetOptions,
    OperationType,
)
from ..exceptions import NotFoundError
from ..model.endpoint import Endpoint
from ..model.node import Node
from ..model.value import SetValueResult, SupervisionResult, Value, get_value_id_str


def get_code_slot_value(node: Node, code_slot: int, property_name: str) -> Value:
    """Get a code slot value."""
    value = node.values.get(
        get_value_id_str(
            node,
            CommandClass.USER_CODE,
            property_name,
            endpoint=0,
            property_key=code_slot,
        )
    )

    if not value:
        raise NotFoundError(f"{property_name} for code slot {code_slot} not found")

    return value


class CodeSlot(TypedDict, total=False):
    """Represent a code slot."""

    code_slot: int  # required
    name: str  # required
    in_use: bool | None  # required
    usercode: str | None


def _get_code_slots(node: Node, include_usercode: bool = False) -> list[CodeSlot]:
    """Get all code slots on the lock and optionally include usercode."""
    code_slot = 1
    slots: list[CodeSlot] = []

    # Loop until we can't find a code slot
    while True:
        try:
            value = get_code_slot_value(node, code_slot, LOCK_USERCODE_PROPERTY)
            status_value = get_code_slot_value(
                node, code_slot, LOCK_USERCODE_STATUS_PROPERTY
            )
        except NotFoundError:
            return slots

        code_slot = int(value.property_key)  # type: ignore[arg-type]
        in_use = (
            None
            if status_value.value is None
            else status_value.value == CodeSlotStatus.ENABLED
        )

        # we know that code slots will always have a property key
        # that is an int, so we can ignore mypy
        slot = {
            ATTR_CODE_SLOT: code_slot,
            ATTR_NAME: value.metadata.label,
            ATTR_IN_USE: in_use,
        }
        if include_usercode:
            slot[ATTR_USERCODE] = value.value

        slots.append(cast(CodeSlot, slot))
        code_slot += 1


def get_code_slots(node: Node) -> list[CodeSlot]:
    """Get all code slots on the lock and whether or not they are used."""
    return _get_code_slots(node, False)


def get_usercodes(node: Node) -> list[CodeSlot]:
    """Get all code slots and usercodes on the lock."""
    return _get_code_slots(node, True)


def get_usercode(node: Node, code_slot: int) -> CodeSlot:
    """Get usercode from slot X on the lock."""
    value = get_code_slot_value(node, code_slot, LOCK_USERCODE_PROPERTY)
    status_value = get_code_slot_value(node, code_slot, LOCK_USERCODE_STATUS_PROPERTY)

    code_slot = int(value.property_key)  # type: ignore[arg-type]
    in_use = (
        None
        if status_value.value is None
        else status_value.value == CodeSlotStatus.ENABLED
    )

    return cast(
        CodeSlot,
        {
            ATTR_CODE_SLOT: code_slot,
            ATTR_NAME: value.metadata.label,
            ATTR_IN_USE: in_use,
            ATTR_USERCODE: value.value,
        },
    )


async def get_usercode_from_node(node: Node, code_slot: int) -> CodeSlot:
    """
    Fetch a usercode directly from a node.

    Should be used when Z-Wave JS's ValueDB hasn't been populated for this code slot.
    This call will populate the ValueDB and trigger value update events from the
    driver.
    """
    # https://zwave-js.github.io/node-zwave-js/#/api/CCs/UserCode?id=get
    await node.async_invoke_cc_api(
        CommandClass.USER_CODE, "get", code_slot, wait_for_result=True
    )
    return get_usercode(node, code_slot)


async def set_usercode(
    node: Node, code_slot: int, usercode: str
) -> SetValueResult | None:
    """Set the usercode to index X on the lock."""
    value = get_code_slot_value(node, code_slot, LOCK_USERCODE_PROPERTY)
    return await node.async_set_value(value, usercode)


async def set_usercodes(node: Node, codes: dict[int, str]) -> SupervisionResult | None:
    """Set the usercode to index X on the lock."""
    cc_api_codes = [
        {
            LOCK_USERCODE_ID_PROPERTY: code_slot,
            LOCK_USERCODE_STATUS_PROPERTY: CodeSlotStatus.ENABLED,
            LOCK_USERCODE_PROPERTY: usercode,
        }
        for code_slot, usercode in codes.items()
    ]
    # https://zwave-js.github.io/node-zwave-js/#/api/CCs/UserCode?id=setmany
    data = await node.async_invoke_cc_api(
        CommandClass.USER_CODE, "setMany", cc_api_codes, wait_for_result=True
    )

    if not data:
        raise ValueError("Received unexpected response from User Code CC setMany API")

    return SupervisionResult(data)


async def clear_usercode(node: Node, code_slot: int) -> SetValueResult | None:
    """Clear a code slot on the lock."""
    value = get_code_slot_value(node, code_slot, LOCK_USERCODE_STATUS_PROPERTY)
    return await node.async_set_value(value, CodeSlotStatus.AVAILABLE)


async def set_configuration(
    endpoint: Endpoint, configuration: DoorLockCCConfigurationSetOptions
) -> SupervisionResult | None:
    """Set lock configuration."""
    # It is invalid to set the operation to timed with no timeout, or to constant
    # with a timeout
    if (configuration.operation_type == OperationType.CONSTANT) ^ (
        configuration.lock_timeout_configuration is None
    ):
        raise ValueError(
            "Invalid operation type and lock timeout configuration combination"
        )
    errors: list[str] = []

    for property_name, attr_name in (
        (CURRENT_AUTO_RELOCK_TIME_PROPERTY, "auto_relock_time"),
        (CURRENT_HOLD_AND_RELEASE_TIME_PROPERTY, "hold_and_release_time"),
        (CURRENT_TWIST_ASSIST_PROPERTY, "twist_assist"),
        (CURRENT_BLOCK_TO_BLOCK_PROPERTY, "block_to_block"),
    ):
        # It a value for a particular configuration value is not provided and it exists
        # on the node, use the cached value
        cached_value = next(
            (
                value
                for value in endpoint.values.values()
                if value.command_class == CommandClass.DOOR_LOCK
                and value.property_name == property_name
            ),
            None,
        )
        if (
            val := getattr(configuration, attr_name)
        ) is not None and cached_value is None:
            errors.append(
                f"- Can't provide value for {property_name} since it is unsupported"
            )
        elif cached_value is not None and val is None and not errors:
            setattr(configuration, attr_name, cached_value.value)

    if errors:
        raise ValueError("\n".join(errors))

    # https://zwave-js.github.io/node-zwave-js/#/api/CCs/UserCode?id=setconfiguration
    data = await endpoint.async_invoke_cc_api(
        CommandClass.DOOR_LOCK, "setConfiguration", configuration.to_dict()
    )

    if not data:
        return None

    return SupervisionResult(data)



"""Constants for the Z-Wave JS python library."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, IntEnum
import logging
from typing import TypedDict

PACKAGE_NAME = "zwave-js-server-python"
__version__ = "0.67.0"

# minimal server schema version we can handle
MIN_SERVER_SCHEMA_VERSION = 44
# max server schema version we can handle (and our code is compatible with)
MAX_SERVER_SCHEMA_VERSION = 44

VALUE_UNKNOWN = "unknown"

NOT_INTERVIEWED = "None"
INTERVIEW_FAILED = "Failed"

CURRENT_STATE_PROPERTY = "currentState"
TARGET_STATE_PROPERTY = "targetState"
CURRENT_VALUE_PROPERTY = "currentValue"
TARGET_VALUE_PROPERTY = "targetValue"
DURATION_PROPERTY = "duration"

TRANSITION_DURATION_OPTION = "transitionDuration"
VOLUME_OPTION = "volume"


class CommandStatus(str, Enum):
    """Status of a command sent to zwave-js-server."""

    ACCEPTED = "accepted"
    QUEUED = "queued"


# Multiple inheritance so that LogLevel will JSON serialize properly
# Reference: https://stackoverflow.com/a/51976841
class LogLevel(str, Enum):
    """Enum for log levels used by node-zwave-js."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/log/shared.ts#L12
    # https://github.com/winstonjs/triple-beam/blame/master/config/npm.js#L14
    ERROR = "error"
    WARN = "warn"
    INFO = "info"
    HTTP = "http"
    VERBOSE = "verbose"
    DEBUG = "debug"
    SILLY = "silly"


LOG_LEVEL_MAP: dict[LogLevel, int] = {
    LogLevel.ERROR: logging.ERROR,
    LogLevel.WARN: logging.WARN,
    LogLevel.INFO: logging.INFO,
    LogLevel.HTTP: logging.INFO,
    LogLevel.VERBOSE: logging.DEBUG,
    LogLevel.DEBUG: logging.DEBUG,
    LogLevel.SILLY: logging.DEBUG,
}


class CommandClass(IntEnum):
    """Enum with all known CommandClasses."""

    SENSOR_ALARM = 156
    SILENCE_ALARM = 157
    SWITCH_ALL = 39
    ANTITHEFT = 93
    ANTITHEFT_UNLOCK = 126
    APPLICATION_CAPABILITY = 87
    APPLICATION_STATUS = 34
    ASSOCIATION = 133
    ASSOCIATION_COMMAND_CONFIGURATION = 155
    ASSOCIATION_GRP_INFO = 89
    AUTHENTICATION = 161
    AUTHENTICATION_MEDIA_WRITE = 162
    BARRIER_OPERATOR = 102
    BASIC = 32
    BASIC_TARIFF_INFO = 54
    BASIC_WINDOW_COVERING = 80
    BATTERY = 128
    SENSOR_BINARY = 48
    SWITCH_BINARY = 37
    SWITCH_TOGGLE_BINARY = 40
    CLIMATE_CONTROL_SCHEDULE = 70
    CENTRAL_SCENE = 91
    CLOCK = 129
    SWITCH_COLOR = 51
    CONFIGURATION = 112
    CONTROLLER_REPLICATION = 33
    CRC_16_ENCAP = 86
    DCP_CONFIG = 58
    DCP_MONITOR = 59
    DEVICE_RESET_LOCALLY = 90
    DOOR_LOCK = 98
    DOOR_LOCK_LOGGING = 76
    ENERGY_PRODUCTION = 144
    ENTRY_CONTROL = 111
    FIRMWARE_UPDATE_MD = 122
    GENERIC_SCHEDULE = 163
    GEOGRAPHIC_LOCATION = 140
    GROUPING_NAME = 123
    HAIL = 130
    HRV_STATUS = 55
    HRV_CONTROL = 57
    HUMIDITY_CONTROL_MODE = 109
    HUMIDITY_CONTROL_OPERATING_STATE = 110
    HUMIDITY_CONTROL_SETPOINT = 100
    INCLUSION_CONTROLLER = 116
    INDICATOR = 135
    IP_ASSOCIATION = 92
    IP_CONFIGURATION = 154
    IR_REPEATER = 160
    IRRIGATION = 107
    LANGUAGE = 137
    LOCK = 118
    MAILBOX = 105
    MANUFACTURER_PROPRIETARY = 145
    MANUFACTURER_SPECIFIC = 114
    MARK = 239
    METER = 50
    METER_TBL_CONFIG = 60
    METER_TBL_MONITOR = 61
    METER_TBL_PUSH = 62
    MTP_WINDOW_COVERING = 81
    MULTI_CHANNEL = 96
    MULTI_CHANNEL_ASSOCIATION = 142
    MULTI_CMD = 143
    SENSOR_MULTILEVEL = 49
    SWITCH_MULTILEVEL = 38
    SWITCH_TOGGLE_MULTILEVEL = 41
    NETWORK_MANAGEMENT_BASIC = 77
    NETWORK_MANAGEMENT_INCLUSION = 52
    NETWORK_MANAGEMENT_INSTALLATION_MAINTENANCE = 103
    NETWORK_MANAGEMENT_PRIMARY = 84
    NETWORK_MANAGEMENT_PROXY = 82
    NO_OPERATION = 0
    NODE_NAMING = 119
    NODE_PROVISIONING = 120
    NOTIFICATION = 113
    POWERLEVEL = 115
    PREPAYMENT = 63
    PREPAYMENT_ENCAPSULATION = 65
    PROPRIETARY = 136
    PROTECTION = 117
    METER_PULSE = 53
    RATE_TBL_CONFIG = 72
    RATE_TBL_MONITOR = 73
    REMOTE_ASSOCIATION_ACTIVATE = 124
    REMOTE_ASSOCIATION = 125
    SCENE_ACTIVATION = 43
    SCENE_ACTUATOR_CONF = 44
    SCENE_CONTROLLER_CONF = 45
    SCHEDULE = 83
    SCHEDULE_ENTRY_LOCK = 78
    SCREEN_ATTRIBUTES = 147
    SCREEN_MD = 146
    SECURITY = 152
    SECURITY_2 = 159
    SECURITY_SCHEME0_MARK = 61696
    SENSOR_CONFIGURATION = 158
    SIMPLE_AV_CONTROL = 148
    SOUND_SWITCH = 121
    SUPERVISION = 108
    TARIFF_CONFIG = 74
    TARIFF_TBL_MONITOR = 75
    THERMOSTAT_FAN_MODE = 68
    THERMOSTAT_FAN_STATE = 69
    THERMOSTAT_MODE = 64
    THERMOSTAT_OPERATING_STATE = 66
    THERMOSTAT_SETBACK = 71
    THERMOSTAT_SETPOINT = 67
    TIME = 138
    TIME_PARAMETERS = 139
    TRANSPORT_SERVICE = 85
    USER_CODE = 99
    VERSION = 134
    WAKE_UP = 132
    WINDOW_COVERING = 106
    ZIP = 35
    ZIP_6LOWPAN = 79
    ZIP_GATEWAY = 95
    ZIP_NAMING = 104
    ZIP_ND = 88
    ZIP_PORTAL = 97
    ZWAVEPLUS_INFO = 94


class ProtocolVersion(IntEnum):
    """Protocol version."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/node/Types.ts#L149
    UNKNOWN = 0
    VERSION_2_0 = 1
    VERSION_4_2X_OR_5_0X = 2
    VERSION_4_5X_OR_6_0X = 3


class NodeStatus(IntEnum):
    """Enum with all Node status values."""

    # https://zwave-js.github.io/node-zwave-js/#/api/node?id=status
    UNKNOWN = 0
    ASLEEP = 1
    AWAKE = 2
    DEAD = 3
    ALIVE = 4


class ConfigurationValueType(str, Enum):
    """Enum for configuration value types."""

    BOOLEAN = "boolean"
    ENUMERATED = "enumerated"
    MANUAL_ENTRY = "manual_entry"
    RANGE = "range"
    UNDEFINED = "undefined"


class NodeType(IntEnum):
    """Enum with all Node types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/capabilities/NodeInfo.ts#L151-L156
    CONTROLLER = 0
    END_NODE = 1


# Exclusion enums
class ExclusionStrategy(IntEnum):
    """Enum with all exclusion strategies."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/controller/Inclusion.ts#L49-L56
    EXCLUDE_ONLY = 0
    DISABLE_PROVISIONING_ENTRY = 1
    UNPROVISION = 2


# Inclusion enums
class InclusionStrategy(IntEnum):
    """Enum for all known inclusion strategies."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/controller/Inclusion.ts#L9-L46
    DEFAULT = 0
    SMART_START = 1
    INSECURE = 2
    SECURITY_S0 = 3
    SECURITY_S2 = 4


class SecurityClass(IntEnum):
    """Enum for all known security classes."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/security/SecurityClass.ts#L3-L17
    NONE = -1
    S2_UNAUTHENTICATED = 0
    S2_AUTHENTICATED = 1
    S2_ACCESS_CONTROL = 2
    S0_LEGACY = 7


class QRCodeVersion(IntEnum):
    """Enum for all known QR Code versions."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/security/QR.ts#L43-L46
    S2 = 0
    SMART_START = 1


class Protocols(IntEnum):
    """Enum for all known protocols."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/capabilities/Protocols.ts#L1-L4
    ZWAVE = 0
    ZWAVE_LONG_RANGE = 1


# https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/security/QR.ts#L41
MINIMUM_QR_STRING_LENGTH = 52


class ZwaveFeature(IntEnum):
    """Enum for all known Zwave features."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/controller/Features.ts#L4
    SMART_START = 0


class PowerLevel(IntEnum):
    """Enum for all known power levels."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/commandclass/PowerlevelCC.ts#L38
    NORMAL_POWER = 0
    DBM_MINUS_1 = 1
    DBM_MINUS_2 = 2
    DBM_MINUS_3 = 3
    DBM_MINUS_4 = 4
    DBM_MINUS_5 = 5
    DBM_MINUS_6 = 6
    DBM_MINUS_7 = 7
    DBM_MINUS_8 = 8
    DBM_MINUS_9 = 9


class InclusionState(IntEnum):
    """Enum for all known inclusion states."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/controller/Inclusion.ts#L154
    IDLE = 0
    INCLUDING = 1
    EXCLUDING = 2
    BUSY = 3
    SMART_START = 4


class RFRegion(IntEnum):
    """Enum for all known RF regions."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/definitions/RFRegion.ts#L1
    EUROPE = 0
    USA = 1
    AUSTRALIA_AND_NEW_ZEALAND = 2
    HONG_KONG = 3
    INDIA = 5
    ISRAEL = 6
    RUSSIA = 7
    CHINA = 8
    USA_LONG_RANGE = 9
    EUROPE_LONG_RANGE = 11
    JAPAN = 32
    KOREA = 33
    UNKNOWN = 254
    DEFAULT_EU = 255


class ProtocolDataRate(IntEnum):
    """Enum for all known protocol data rates."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/capabilities/Protocols.ts#L6

    ZWAVE_9K6 = 1
    ZWAVE_40K = 2
    ZWAVE_100K = 3
    LONG_RANGE_100K = 4


class RssiError(IntEnum):
    """Enum for all known RSSI errors."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/controller/SendDataShared.ts#L79
    NOT_AVAILABLE = 127
    RECEIVER_SATURATED = 126
    NO_SIGNAL_DETECTED = 125


class ProvisioningEntryStatus(IntEnum):
    """Enum for all known provisioning entry statuses."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/controller/Inclusion.ts#L136
    ACTIVE = 0
    INACTIVE = 1


class SecurityBootstrapFailure(IntEnum):
    """Enum with all security bootstrap failure reasons."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/controller/Inclusion.ts#L16
    USER_CANCELED = 0
    NO_KEYS_CONFIGURED = 1
    S2_NO_USER_CALLBACKS = 2
    TIMEOUT = 3
    PARAMETER_MISMATCH = 4
    NODE_CANCELED = 5
    S2_INCORRECT_PIN = 6
    S2_WRONG_SECURITY_LEVEL = 7
    UNKNOWN = 8


class SetValueStatus(IntEnum):
    """Enum for all known setValue statuses."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/cc/src/lib/API.ts#L83
    # The device reports no support for this command
    NO_DEVICE_SUPPORT = 0
    # The device has accepted the command and is working on it
    WORKING = 1
    # The device has rejected the command
    FAIL = 2
    # The endpoint specified in the value ID does not exist
    ENDPOINT_NOT_FOUND = 3
    # The given CC or its API is not implemented (yet) or it has no `setValue`
    # implementation
    NOT_IMPLEMENTED = 4
    # The value to set (or a related value) is invalid
    INVALID_VALUE = 5
    # The command was sent successfully, but it is unknown whether it was executed
    SUCCESS_UNSUPERVISED = 254
    # The device has executed the command successfully
    SUCCESS = 255


SET_VALUE_SUCCESS = (
    SetValueStatus.SUCCESS,
    SetValueStatus.SUCCESS_UNSUPERVISED,
    SetValueStatus.WORKING,
)


class RemoveNodeReason(IntEnum):
    """Enum for all known reasons why a node was removed."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/controller/Inclusion.ts#L266
    # The node was excluded by the user or an inclusion controller
    EXCLUDED = 0
    # The node was excluded by an inclusion controller
    PROXY_EXCLUDED = 1
    # The node was removed using the "remove failed node" feature
    REMOVE_FAILED = 2
    # The node was replaced using the "replace failed node" feature
    REPLACED = 3
    # The node was replaced by an inclusion controller
    PROXY_REPLACED = 4
    # The node was reset locally and was auto-removed
    RESET = 5
    # SmartStart inclusion failed, and the node was auto-removed as a result.
    SMART_START_FAILED = 6


class Weekday(IntEnum):
    """Enum for all known weekdays."""

    UNKNOWN = 0
    MONDAY = 1
    TUESDAY = 2
    WEDNESDAY = 3
    THURSDAY = 4
    FRIDAY = 5
    SATURDAY = 6
    SUNDAY = 7


class DateAndTimeDataType(TypedDict, total=False):
    """Represent a date and time data type."""

    hour: int
    minute: int
    weekday: int
    second: int
    year: int
    month: int
    day: int
    dstOffset: int
    standardOffset: int


@dataclass
class DateAndTime:
    """Represent a date and time."""

    data: DateAndTimeDataType = field(repr=False)
    hour: int | None = field(init=False)
    minute: int | None = field(init=False)
    weekday: Weekday | None = field(default=None, init=False)
    second: int | None = field(init=False)
    year: int | None = field(init=False)
    month: int | None = field(init=False)
    day: int | None = field(init=False)
    dst_offset: int | None = field(init=False)
    standard_offset: int | None = field(init=False)

    def __post_init__(self) -> None:
        """Post initialization."""
        self.hour = self.data.get("hour")
        self.minute = self.data.get("minute")
        if weekday := self.data.get("weekday"):
            self.weekday = Weekday(weekday)
        self.second = self.data.get("second")
        self.year = self.data.get("year")
        self.month = self.data.get("month")
        self.day = self.data.get("day")
        self.dst_offset = self.data.get("dstOffset")
        self.standard_offset = self.data.get("standardOffset")


class ControllerStatus(IntEnum):
    """Enum for all known controller statuses."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/consts/ControllerStatus.TS
    # The controller is ready to accept commands and transmit
    READY = 0
    # The controller is unresponsive
    UNRESPONSIVE = 1
    # The controller is unable to transmit
    JAMMED = 2


class SupervisionStatus(IntEnum):
    """Enum for all known supervision statuses."""

    # https://github.com/zwave-js/node-zwave-js/blob/cc_api_options/packages/core/src/consts/Transmission.ts#L304
    NO_SUPPORT = 0
    WORKING = 1
    FAIL = 2
    SUCCESS = 255


class AssociationCheckResult(IntEnum):
    """Enum for all known association check results."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/cc/src/lib/_Types.ts#L47
    OK = 1
    FORBIDDEN_DESTINATION_IS_LONG_RANGE = 2
    FORBIDDEN_SOURCE_IS_LONG_RANGE = 3
    FORBIDDEN_SELF_ASSOCIATION = 4
    FORBIDDEN_SECURITY_CLASS_MISMATCH = 5
    FORBIDDEN_DESTINATION_SECURITY_CLASS_NOT_GRANTED = 6
    FORBIDDEN_NO_SUPPORTED_CCS = 7



"""Provide a model for the log config."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TypedDict, cast

from ..const import LogLevel


class LogConfigDataType(TypedDict, total=False):
    """Represent a log config data dict type."""

    enabled: bool
    level: str
    logToFile: bool
    filename: str
    forceConsole: bool


@dataclass
class LogConfig:
    """Represent a log config dict type."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/core/src/log/shared.ts#L85
    # Must include at least one key
    enabled: bool | None = None
    level: LogLevel | None = None
    log_to_file: bool | None = None
    filename: str | None = None
    force_console: bool | None = None

    def to_dict(self) -> LogConfigDataType:
        """Return LogConfigDataType dict from self."""
        data = {
            "enabled": self.enabled,
            "level": self.level.value if self.level else None,
            "logToFile": self.log_to_file,
            "filename": self.filename,
            "forceConsole": self.force_console,
        }
        return cast(LogConfigDataType, {k: v for k, v in data.items() if v is not None})

    @classmethod
    def from_dict(cls, data: LogConfigDataType) -> LogConfig:
        """Return LogConfig from LogConfigDataType dict."""
        return cls(
            data.get("enabled"),
            LogLevel(data["level"]) if "level" in data else None,
            data.get("logToFile"),
            data.get("filename"),
            data.get("forceConsole"),
        )



"""Constants for the Multilevel Switch CC."""

from __future__ import annotations

from enum import IntEnum

SET_TO_PREVIOUS_VALUE = 255

COVER_OPEN_PROPERTY = "Open"
COVER_UP_PROPERTY = "Up"
COVER_ON_PROPERTY = "On"
COVER_CLOSE_PROPERTY = "Close"
COVER_DOWN_PROPERTY = "Down"
COVER_OFF_PROPERTY = "Off"


class CoverStates(IntEnum):
    """Enum with all (known/used) Z-Wave Cover States."""

    CLOSED = 0
    OPEN = 99


class MultilevelSwitchCommand(IntEnum):
    """Enum for known multilevel switch notifications."""

    START_LEVEL_CHANGE = 4
    STOP_LEVEL_CHANGE = 5



"""Provide models for rebuilding routes."""

from __future__ import annotations

from dataclasses import dataclass
from enum import StrEnum
from typing import TypedDict


class RebuildRoutesOptionsDataType(TypedDict, total=False):
    """Represent a rebuild routes options data dict type."""

    includeSleeping: bool


@dataclass
class RebuildRoutesOptions:
    """Represent options for rebuilding routes."""

    include_sleeping: bool | None = None

    @classmethod
    def from_dict(cls, data: RebuildRoutesOptionsDataType) -> RebuildRoutesOptions:
        """Return options from data."""
        return cls(include_sleeping=data.get("includeSleeping"))

    def to_dict(self) -> RebuildRoutesOptionsDataType:
        """Return dict representation of data."""
        if self.include_sleeping is None:
            return {}
        return {"includeSleeping": self.include_sleeping}


class RebuildRoutesStatus(StrEnum):
    """Enum of all known rebuild routes status values."""

    PENDING = "pending"
    DONE = "done"
    FAILED = "failed"
    SKIPPED = "skipped"



"""Test the client."""

import asyncio
from datetime import datetime
import logging
from unittest.mock import Mock, patch

from aiohttp.client_exceptions import ClientError, WSServerHandshakeError
from aiohttp.client_reqrep import ClientResponse, RequestInfo
from aiohttp.http_websocket import WSMsgType
import pytest

from test.common import MockCommandProtocol
from zwave_js_server.client import LOGGER, Client
from zwave_js_server.const import MAX_SERVER_SCHEMA_VERSION, LogLevel, __version__
from zwave_js_server.event import Event
from zwave_js_server.exceptions import (
    CannotConnect,
    ConnectionFailed,
    FailedCommand,
    FailedZWaveCommand,
    InvalidMessage,
    InvalidServerVersion,
    InvalidState,
    NotConnected,
)
from zwave_js_server.model.driver import Driver
from zwave_js_server.model.log_config import LogConfig


async def test_connect_disconnect(client_session, url):
    """Test client connect and disconnect."""
    async with Client(url, client_session) as client:
        assert client.connected

    assert not client.connected


@pytest.mark.parametrize(
    "error",
    [ClientError, WSServerHandshakeError(Mock(RequestInfo), (Mock(ClientResponse),))],
)
async def test_cannot_connect(client_session, url, error):
    """Test cannot connect."""
    client_session.ws_connect.side_effect = error
    client = Client(url, client_session)

    with pytest.raises(CannotConnect):
        await client.connect()

    assert not client.connected


async def test_send_command_schema(
    client_session, url, ws_client, driver_ready, driver
):
    """Test sending unsupported command."""
    client = Client(url, client_session)
    await client.connect()
    assert client.connected
    client.driver = driver
    await client.listen(driver_ready)
    ws_client.receive.assert_awaited()

    # test schema version is at server maximum
    if client.version.max_schema_version < MAX_SERVER_SCHEMA_VERSION:
        assert client.schema_version == client.version.max_schema_version

    # send command of current schema version should not fail
    with pytest.raises(NotConnected):
        await client.async_send_command(
            {"command": "test"}, require_schema=client.schema_version
        )
    # send command of unsupported schema version should fail
    with pytest.raises(InvalidServerVersion):
        await client.async_send_command(
            {"command": "test"}, require_schema=client.schema_version + 2
        )
    with pytest.raises(InvalidServerVersion):
        await client.async_send_command_no_wait(
            {"command": "test"}, require_schema=client.schema_version + 2
        )


async def test_min_schema_version(client_session, url, version_data):
    """Test client connect with invalid schema version."""
    version_data["minSchemaVersion"] = 100
    client = Client(url, client_session)

    with pytest.raises(InvalidServerVersion):
        await client.connect()

    assert not client.connected


async def test_max_schema_version(client_session, url, version_data):
    """Test client connect with invalid schema version."""
    version_data["maxSchemaVersion"] = 0
    client = Client(url, client_session)

    with pytest.raises(InvalidServerVersion):
        await client.connect()

    assert not client.connected


async def test_send_json_when_disconnected(client_session, url):
    """Test send json message when disconnected."""
    client = Client(url, client_session)

    assert not client.connected

    with pytest.raises(NotConnected):
        await client.async_send_command({"test": None})


async def test_listen(client_session, url, driver_ready):
    """Test client listen."""
    client = Client(url, client_session)

    assert not client.driver

    await client.connect()

    assert client.connected

    asyncio.create_task(client.listen(driver_ready))
    await driver_ready.wait()
    assert client.driver

    await client.disconnect()
    assert not client.connected


async def test_listen_client_error(
    client_session, url, ws_client, messages, ws_message, driver_ready
):
    """Test websocket error on listen."""
    client = Client(url, client_session)
    await client.connect()
    assert client.connected

    messages.append(ws_message)

    ws_client.receive.side_effect = asyncio.CancelledError()

    # This should break out of the listen loop before any message is received.
    with pytest.raises(asyncio.CancelledError):
        await client.listen(driver_ready)

    assert not ws_message.json.called


@pytest.mark.parametrize(
    "message_type, exception",
    [
        (WSMsgType.ERROR, ConnectionFailed),
        (WSMsgType.BINARY, InvalidMessage),
    ],
)
async def test_listen_error_message_types(
    client_session, url, messages, ws_message, message_type, exception, driver_ready
):
    """Test different websocket message types that should raise on listen."""
    client = Client(url, client_session)
    await client.connect()
    assert client.connected

    ws_message.type = message_type
    messages.append(ws_message)

    with pytest.raises(exception):
        await client.listen(driver_ready)


@pytest.mark.parametrize(
    "message_type", [WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING]
)
async def test_listen_disconnect_message_types(
    client_session, url, ws_client, messages, ws_message, message_type, driver_ready
):
    """Test different websocket message types that stop listen."""
    async with Client(url, client_session) as client:
        assert client.connected
        ws_message.type = message_type
        messages.append(ws_message)

        # This should break out of the listen loop before handling the received message.
        # Otherwise there will be an error.
        await client.listen(driver_ready)

    # Assert that we received a message.
    ws_client.receive.assert_awaited()


async def test_listen_invalid_message_data(
    client_session, url, messages, ws_message, driver_ready
):
    """Test websocket message data that should raise on listen."""
    client = Client(url, client_session)
    await client.connect()
    assert client.connected

    ws_message.json.side_effect = ValueError("Boom")
    messages.append(ws_message)

    with pytest.raises(InvalidMessage):
        await client.listen(driver_ready)


async def test_listen_not_success(client_session, url, result, driver_ready):
    """Test receive result message with success False on listen."""
    result["success"] = False
    result["errorCode"] = "error_code"
    result["message"] = "test"
    client = Client(url, client_session)
    await client.connect()

    with pytest.raises(FailedCommand):
        await client.listen(driver_ready)

    assert not client.connected


async def test_initialize_not_success(
    client_session, url, initialize_data, driver_ready
):
    """Test receive result message with success False on listen."""
    initialize_data["success"] = False
    initialize_data["errorCode"] = "error_code"
    initialize_data["message"] = "test"
    client = Client(url, client_session)
    await client.connect()

    with pytest.raises(FailedCommand):
        await client.listen(driver_ready)

    assert not client.connected


async def test_get_log_config_not_success(
    client_session, url, get_log_config_data, driver_ready
):
    """Test receive log config message with success False on listen."""
    get_log_config_data["success"] = False
    get_log_config_data["errorCode"] = "error_code"
    get_log_config_data["message"] = "test"
    client = Client(url, client_session)
    await client.connect()

    with pytest.raises(FailedCommand):
        await client.listen(driver_ready)

    assert not client.connected


async def test_listen_without_connect(client_session, url, driver_ready):
    """Test listen without first being connected."""
    client = Client(url, client_session)
    assert not client.connected

    with pytest.raises(InvalidState):
        await client.listen(driver_ready)


async def test_listen_event(
    client_session, url, ws_client, messages, ws_message, result, driver_ready
):
    """Test receiving event result type on listen."""
    client = Client(url, client_session)
    await client.connect()

    assert client.connected

    result["type"] = "event"
    result["event"] = {
        "source": "node",
        "event": "value updated",
        "nodeId": 52,
        "args": {
            "commandClassName": "Basic",
            "commandClass": 32,
            "endpoint": 0,
            "property": "currentValue",
            "newValue": 255,
            "prevValue": 255,
            "propertyName": "currentValue",
        },
    }
    messages.append(ws_message)

    await client.listen(driver_ready)
    ws_client.receive.assert_awaited()


async def test_listen_unknown_result_type(
    client_session, url, ws_client, result, driver_ready, driver
):
    """Test websocket message with unknown type on listen."""
    client = Client(url, client_session)
    await client.connect()

    assert client.connected

    # Make sure there's a driver so we can test an unknown event.
    client.driver = driver
    result["type"] = "unknown"

    # Receiving an unknown message type should not error.
    await client.listen(driver_ready)

    ws_client.receive.assert_awaited()


async def test_command_error_handling(client, mock_command):
    """Test error handling."""
    mock_command(
        {"command": "some_command"},
        {"errorCode": "unknown_command", "message": "test"},
        False,
    )

    with pytest.raises(FailedCommand) as raised:
        await client.async_send_command({"command": "some_command"})

    assert raised.value.error_code == "unknown_command"
    assert str(raised.value) == "unknown_command: test"

    mock_command(
        {"command": "some_zjs_command"},
        {
            "errorCode": "zwave_error",
            "zwaveErrorCode": 3,
            "zwaveErrorMessage": "Node 5 is dead",
        },
        False,
    )

    with pytest.raises(FailedZWaveCommand) as raised:
        await client.async_send_command({"command": "some_zjs_command"})

    assert raised.value.error_code == "zwave_error"
    assert raised.value.zwave_error_code == 3
    assert raised.value.zwave_error_message == "Node 5 is dead"


async def test_record_messages(client, wallmote_central_scene, mock_command, uuid4):
    """Test recording messages."""
    # pylint: disable=protected-access
    assert not client.recording_messages
    assert not client._recorded_commands
    assert not client._recorded_events
    client.begin_recording_messages()
    mock_command(
        {"command": "some_command"},
        {},
    )

    with pytest.raises(InvalidState):
        client.begin_recording_messages()

    with patch("zwave_js_server.client.datetime") as mock_dt:
        mock_dt.utcnow.return_value = datetime(2022, 1, 7, 1)

        await client.async_send_command({"command": "some_command"})

    assert len(client._recorded_commands) == 1
    assert len(client._recorded_events) == 0
    assert uuid4 in client._recorded_commands
    assert client._recorded_commands[uuid4]["record_type"] == "command"
    assert client._recorded_commands[uuid4]["command"] == "some_command"
    assert client._recorded_commands[uuid4]["command_msg"] == {
        "command": "some_command",
        "messageId": uuid4,
    }
    assert client._recorded_commands[uuid4]["result_msg"] == {
        "messageId": "1234",
        "result": {},
        "success": True,
        "type": "result",
    }
    assert "ts" in client._recorded_commands[uuid4]
    assert "result_ts" in client._recorded_commands[uuid4]

    with patch("zwave_js_server.client.datetime") as mock_dt:
        mock_dt.utcnow.return_value = datetime(2022, 1, 7, 0)
        client._handle_incoming_message(
            {
                "type": "event",
                "event": {
                    "source": "node",
                    "event": "value updated",
                    "nodeId": wallmote_central_scene.node_id,
                    "args": {
                        "commandClassName": "Binary Switch",
                        "commandClass": 39,
                        "endpoint": 0,
                        "property": "currentValue",
                        "newValue": False,
                        "prevValue": True,
                        "propertyName": "currentValue",
                    },
                },
            }
        )
    assert len(client._recorded_commands) == 1
    assert len(client._recorded_events) == 1
    logging.getLogger(__name__).error(client._recorded_events)
    event = client._recorded_events[0]
    assert event["record_type"] == "event"
    assert event["type"] == "value updated"
    assert event["event_msg"] == {
        "type": "event",
        "event": {
            "source": "node",
            "event": "value updated",
            "nodeId": wallmote_central_scene.node_id,
            "args": {
                "commandClassName": "Binary Switch",
                "commandClass": 39,
                "endpoint": 0,
                "property": "currentValue",
                "newValue": False,
                "prevValue": True,
                "propertyName": "currentValue",
            },
        },
    }
    assert "ts" in event

    replay_dump = client.end_recording_messages()

    assert len(replay_dump) == 2
    assert len(client._recorded_commands) == 0
    assert len(client._recorded_events) == 0

    # Testing that events are properly sorted by timestamp. Even though the event
    # comes after the command in the code, the patch should make the event appear first
    assert replay_dump[0]["record_type"] == "event"

    with pytest.raises(InvalidState):
        client.end_recording_messages()


async def test_additional_user_agent_components(client_session, url):
    """Test additionalUserAgentComponents parameter."""
    # pylint: disable=protected-access
    client = Client(
        url, client_session, additional_user_agent_components={"foo": "bar"}
    )
    client._client = True
    with (
        patch(
            "zwave_js_server.client.Client._send_json_message", return_value=None
        ) as send_json_mock,
        patch(
            "zwave_js_server.client.Client._receive_json_or_raise",
            return_value={"success": True},
        ),
    ):
        await client.initialize()
        send_json_mock.assert_called_once_with(
            {
                "command": "initialize",
                "messageId": "initialize",
                "schemaVersion": 44,
                "additionalUserAgentComponents": {
                    "zwave-js-server-python": __version__,
                    "foo": "bar",
                },
            }
        )


async def test_pop_future_none(client_session, url, driver_ready):
    """Test when a future has been cleared from futures dict, popping still works."""
    client = Client(url, client_session)
    await client.connect()

    assert client.connected

    asyncio.create_task(client.listen(driver_ready))

    await driver_ready.wait()

    with pytest.raises(asyncio.CancelledError):
        await client.async_send_command({"command": "some_command"})


async def test_log_server(
    client: Client,
    driver: Driver,
    caplog: pytest.LogCaptureFixture,
    mock_command: MockCommandProtocol,
) -> None:
    """Test logging from server."""
    # pylint: disable=protected-access
    assert client.connected
    mock_command(
        {"command": "start_listening_logs"},
        {},
    )
    mock_command(
        {"command": "stop_listening_logs"},
        {},
    )
    # Set log levels to force the lib to change log levels
    LOGGER.setLevel(logging.INFO)
    client.driver.log_config = LogConfig(True, LogLevel.DEBUG, False, None, None)
    await client.enable_server_logging()
    assert client.server_logging_enabled
    assert len(caplog.records) == 2
    assert "logging is currently more verbose" in caplog.records[0].message
    assert caplog.records[0].name == "zwave_js_server"

    # Test that enabling again is a no-op
    await client.enable_server_logging()
    assert client.server_logging_enabled
    assert len(caplog.records) == 2

    LOGGER.setLevel(logging.INFO)
    event = Event(
        "log config updated",
        data={
            "source": "driver",
            "event": "log config updated",
            "config": {"level": "silly"},
        },
    )
    driver.receive_event(event)

    assert len(caplog.records) == 3
    assert "logging is currently more verbose" in caplog.records[2].message
    assert caplog.records[2].name == "zwave_js_server"

    event = Event(
        type="logging",
        data={
            "source": "driver",
            "event": "logging",
            "formattedMessage": [
                "2021-04-18T18:03:34.051Z CNTRLR   [Node 005] [~] \n",
                "test",
            ],
            "level": "debug",
            "primaryTags": "[Node 005] [~] [Notification]",
            "secondaryTags": "[Endpoint 0]",
            "message": "Home Security[Motion sensor status]\n: 8 => 0",
            "direction": "  ",
            "label": "CNTRLR",
            "timestamp": "2021-04-18T18:03:34.051Z",
            "multiline": True,
            "secondaryTagPadding": -1,
            "context": {
                "source": "controller",
                "type": "node",
                "nodeId": 5,
                "header": "Notification",
                "direction": "none",
                "change": "notification",
                "endpoint": 0,
            },
        },
    )
    driver.receive_event(event)
    assert len(caplog.records) == 4
    assert "Node 005" in caplog.records[3].message
    assert caplog.records[3].levelno == logging.DEBUG
    assert caplog.records[3].name == "zwave_js_server.server"

    # First time we disable should be clean
    client.disable_server_logging()
    assert not client.server_logging_enabled
    assert len(caplog.records) == 4

    # Test that disabling again is a no-op
    client.disable_server_logging()
    assert not client.server_logging_enabled

    # Test that enabling server logging raises an error when client is not connected
    client.driver = None
    client._client = None

    with pytest.raises(InvalidState):
        await client.enable_server_logging()

    client.disable_server_logging()
